"""
Compendium of generic DNS utilities
# Examples:
dns.lookup(name, rdtype, ...)
dns.query(name, rdtype, ...)

dns.srv_rec(data)
dns.srv_data('my1.example.com', 389, prio=10, weight=100)
dns.srv_name('ldap/tcp', 'example.com')

"""

import base64
import binascii
import functools
import hashlib
import itertools
import logging
import random
import re
import shlex
import socket
import ssl
import string

import salt.modules.cmdmod
import salt.utils.files
import salt.utils.network
import salt.utils.path
import salt.utils.stringutils
from salt._compat import ipaddress
from salt.utils.odict import OrderedDict

# Integrations
try:
    import dns.resolver  # pylint: disable=no-name-in-module

    HAS_DNSPYTHON = True
except ImportError:
    HAS_DNSPYTHON = False
try:
    import tldextract

    HAS_TLDEXTRACT = True
except ImportError:
    HAS_TLDEXTRACT = False
HAS_DIG = salt.utils.path.which("dig") is not None
DIG_OPTIONS = "+search +fail +noall +answer +nocl +nottl"
HAS_DRILL = salt.utils.path.which("drill") is not None
HAS_HOST = salt.utils.path.which("host") is not None
HAS_NSLOOKUP = salt.utils.path.which("nslookup") is not None

__salt__ = {"cmd.run_all": salt.modules.cmdmod.run_all}
log = logging.getLogger(__name__)


class RFC:
    """
    Simple holding class for all RFC/IANA registered lists & standards
    """

    # https://tools.ietf.org/html/rfc6844#section-3
    CAA_TAGS = ("issue", "issuewild", "iodef")

    # http://www.iana.org/assignments/dns-sshfp-rr-parameters/dns-sshfp-rr-parameters.xhtml
    SSHFP_ALGO = OrderedDict(
        (
            (1, "rsa"),
            (2, "dsa"),
            (3, "ecdsa"),
            (4, "ed25519"),
        )
    )

    SSHFP_HASH = OrderedDict(
        (
            (1, "sha1"),
            (2, "sha256"),
        )
    )

    # http://www.iana.org/assignments/dane-parameters/dane-parameters.xhtml
    TLSA_USAGE = OrderedDict(
        (
            (0, "pkixta"),
            (1, "pkixee"),
            (2, "daneta"),
            (3, "daneee"),
        )
    )

    TLSA_SELECT = OrderedDict(
        (
            (0, "cert"),
            (1, "spki"),
        )
    )

    TLSA_MATCHING = OrderedDict(
        (
            (0, "full"),
            (1, "sha256"),
            (2, "sha512"),
        )
    )

    SRV_PROTO = ("tcp", "udp", "sctp")

    @staticmethod
    def validate(lookup, ref, match=None):
        if lookup in ref:
            return lookup
        elif match == "in":
            return [code for code, name in ref.items() if lookup in name][-1]
        else:
            # OrderedDicts only!(?)
            return {name: code for code, name in ref.items()}[lookup]


def _to_port(port):
    try:
        port = int(port)
        assert 1 <= port <= 65535
        return port
    except (ValueError, AssertionError):
        raise ValueError(f"Invalid port {port}")


def _tree(domain, tld=False):
    """
    Split out a domain in its parents

    Leverages tldextract to take the TLDs from publicsuffix.org
    or makes a valiant approximation of that

    :param domain: dc2.ams2.example.com
    :param tld: Include TLD in list
    :return: [ 'dc2.ams2.example.com', 'ams2.example.com', 'example.com']
    """
    domain = domain.rstrip(".")
    assert "." in domain, "Provide a decent domain"

    if not tld:
        if HAS_TLDEXTRACT:
            tld = tldextract.extract(domain).suffix
        else:
            tld = re.search(
                r"((?:(?:ac|biz|com?|info|edu|gov|mil|name|net|n[oi]m|org)\.)?[^.]+)$",
                domain,
            ).group()
            log.info(
                "Without tldextract, dns.util resolves the TLD of %s to %s", domain, tld
            )

    res = [domain]
    while True:
        idx = domain.find(".")
        if idx < 0:
            break
        domain = domain[idx + 1 :]
        if domain == tld:
            break
        res.append(domain)

    return res


def _weighted_order(recs):
    res = []
    weights = [rec["weight"] for rec in recs]
    while weights:
        rnd = random.random() * sum(weights)
        for i, w in enumerate(weights):
            rnd -= w
            if rnd < 0:
                res.append(recs.pop(i)["name"])
                weights.pop(i)
                break

    return res


def _cast(rec_data, rec_cast):
    if isinstance(rec_cast, dict):
        rec_data = type(next(iter(rec_cast.keys())))(rec_data)
        res = rec_cast[rec_data]
        return res
    elif isinstance(rec_cast, (list, tuple)):
        return RFC.validate(rec_data, rec_cast)
    else:
        return rec_cast(rec_data)


def _data2rec(schema, rec_data):
    """
    schema = OrderedDict({
        'prio': int,
        'weight': int,
        'port': to_port,
        'name': str,
    })
    rec_data = '10 20 25 myawesome.nl'

    res = {'prio': 10, 'weight': 20, 'port': 25 'name': 'myawesome.nl'}
    """
    try:
        rec_fields = rec_data.split(" ")
        # spaces in digest fields are allowed
        assert len(rec_fields) >= len(schema)
        if len(rec_fields) > len(schema):
            cutoff = len(schema) - 1
            rec_fields = rec_fields[0:cutoff] + ["".join(rec_fields[cutoff:])]

        if len(schema) == 1:
            res = _cast(rec_fields[0], next(iter(schema.values())))
        else:
            res = {
                field_name: _cast(rec_field, rec_cast)
                for (field_name, rec_cast), rec_field in zip(schema.items(), rec_fields)
            }
        return res
    except (AssertionError, AttributeError, TypeError, ValueError) as e:
        raise ValueError(
            'Unable to cast "{0}" as "{2}": {1}'.format(
                rec_data, e, " ".join(schema.keys())
            )
        )


def _data2rec_group(schema, recs_data, group_key):
    if not isinstance(recs_data, (list, tuple)):
        recs_data = [recs_data]

    res = OrderedDict()

    try:
        for rdata in recs_data:
            rdata = _data2rec(schema, rdata)
            assert rdata and group_key in rdata

            idx = rdata.pop(group_key)
            if idx not in res:
                res[idx] = []

            if len(rdata) == 1:
                rdata = next(iter(rdata.values()))

            res[idx].append(rdata)
        return res
    except (AssertionError, ValueError) as e:
        raise ValueError(
            'Unable to cast "{}" as a group of "{}": {}'.format(
                ",".join(recs_data), " ".join(schema.keys()), e
            )
        )


def _rec2data(*rdata):
    return " ".join(rdata)


def _data_clean(data):
    data = data.strip(string.whitespace)
    if data.startswith(('"', "'")) and data.endswith(('"', "'")):
        return data[1:-1]
    else:
        return data


def _lookup_dig(name, rdtype, timeout=None, servers=None, secure=None):
    """
    Use dig to lookup addresses
    :param name: Name of record to search
    :param rdtype: DNS record type
    :param timeout: server response timeout
    :param servers: [] of servers to use
    :return: [] of records or False if error
    """
    cmd = f"dig {DIG_OPTIONS} -t {rdtype} "
    if servers:
        cmd += "".join([f"@{srv} " for srv in servers])
    if timeout is not None:
        if servers:
            timeout = int(float(timeout) / len(servers))
        else:
            timeout = int(timeout)
        cmd += f"+time={timeout} "
    if secure:
        cmd += "+dnssec +adflag "

    cmd = __salt__["cmd.run_all"](
        f"{cmd} {name}", python_shell=False, output_loglevel="quiet"
    )

    if "ignoring invalid type" in cmd["stderr"]:
        raise ValueError(f"Invalid DNS type {rdtype}")
    elif cmd["retcode"] != 0:
        log.warning(
            "dig returned (%s): %s",
            cmd["retcode"],
            cmd["stderr"].strip(string.whitespace + ";"),
        )
        return False
    elif not cmd["stdout"]:
        return []

    validated = False
    res = []
    for line in cmd["stdout"].splitlines():
        _, rtype, rdata = line.split(None, 2)
        if rtype == "CNAME" and rdtype != "CNAME":
            continue
        elif rtype == "RRSIG":
            validated = True
            continue
        res.append(_data_clean(rdata))

    if res and secure and not validated:
        return False
    else:
        return res


def _lookup_drill(name, rdtype, timeout=None, servers=None, secure=None):
    """
    Use drill to lookup addresses
    :param name: Name of record to search
    :param rdtype: DNS record type
    :param timeout: command return timeout
    :param servers: [] of servers to use
    :return: [] of records or False if error
    """
    cmd = "drill "
    if secure:
        cmd += "-D -o ad "
    cmd += f"{rdtype} {name} "
    if servers:
        cmd += "".join([f"@{srv} " for srv in servers])
    cmd = __salt__["cmd.run_all"](
        cmd, timeout=timeout, python_shell=False, output_loglevel="quiet"
    )

    if cmd["retcode"] != 0:
        log.warning("drill returned (%s): %s", cmd["retcode"], cmd["stderr"])
        return False

    lookup_res = iter(cmd["stdout"].splitlines())
    validated = False
    res = []
    try:
        line = ""
        while "ANSWER SECTION" not in line:
            line = next(lookup_res)
        while True:
            line = next(lookup_res)
            line = line.strip()
            if not line or line.startswith(";;"):
                break

            l_type, l_rec = line.split(None, 4)[-2:]
            if l_type == "CNAME" and rdtype != "CNAME":
                continue
            elif l_type == "RRSIG":
                validated = True
                continue
            elif l_type != rdtype:
                raise ValueError(f"Invalid DNS type {rdtype}")

            res.append(_data_clean(l_rec))

    except StopIteration:
        pass

    if res and secure and not validated:
        return False
    else:
        return res


def _lookup_gai(name, rdtype, timeout=None):
    """
    Use Python's socket interface to lookup addresses
    :param name: Name of record to search
    :param rdtype: A or AAAA
    :param timeout: ignored
    :return: [] of addresses or False if error
    """
    if rdtype == "A":
        sock_t = socket.AF_INET
    elif rdtype == "AAAA":
        sock_t = socket.AF_INET6
    else:
        raise ValueError(f"Invalid DNS type {rdtype} for gai lookup")

    if timeout:
        log.info("Ignoring timeout on gai resolver; fix resolv.conf to do that")

    try:
        addresses = [
            sock[4][0]
            for sock in socket.getaddrinfo(name, None, sock_t, 0, socket.SOCK_RAW)
        ]
        return addresses
    except socket.gaierror:
        return False


def _lookup_host(name, rdtype, timeout=None, server=None):
    """
    Use host to lookup addresses
    :param name: Name of record to search
    :param server: Server to query
    :param rdtype: DNS record type
    :param timeout: server response wait
    :return: [] of records or False if error
    """
    cmd = f"host -t {rdtype} "

    if timeout:
        cmd += f"-W {int(timeout)} "
    cmd += name
    if server is not None:
        cmd += f" {server}"

    cmd = __salt__["cmd.run_all"](cmd, python_shell=False, output_loglevel="quiet")

    if "invalid type" in cmd["stderr"]:
        raise ValueError(f"Invalid DNS type {rdtype}")
    elif cmd["retcode"] != 0:
        log.warning("host returned (%s): %s", cmd["retcode"], cmd["stderr"])
        return False
    elif "has no" in cmd["stdout"]:
        return []

    res = []
    _stdout = cmd["stdout"] if server is None else cmd["stdout"].split("\n\n")[-1]
    for line in _stdout.splitlines():
        if rdtype != "CNAME" and "is an alias" in line:
            continue
        line = line.split(" ", 3)[-1]
        for prefix in ("record", "address", "handled by", "alias for"):
            if line.startswith(prefix):
                line = line[len(prefix) + 1 :]
                break
        res.append(_data_clean(line))

    return res


def _lookup_dnspython(name, rdtype, timeout=None, servers=None, secure=None):
    """
    Use dnspython to lookup addresses
    :param name: Name of record to search
    :param rdtype: DNS record type
    :param timeout: query timeout
    :param server: [] of server(s) to try in order
    :return: [] of records or False if error
    """
    resolver = dns.resolver.Resolver()

    if timeout is not None:
        resolver.lifetime = float(timeout)
    if servers:
        resolver.nameservers = servers
    if secure:
        resolver.ednsflags += dns.flags.DO

    try:
        res = [
            _data_clean(rr.to_text())
            for rr in resolver.query(name, rdtype, raise_on_no_answer=False)
        ]
        return res
    except dns.rdatatype.UnknownRdatatype:
        raise ValueError(f"Invalid DNS type {rdtype}")
    except (
        dns.resolver.NXDOMAIN,
        dns.resolver.YXDOMAIN,
        dns.resolver.NoNameservers,
        dns.exception.Timeout,
    ):
        return False


def _lookup_nslookup(name, rdtype, timeout=None, server=None):
    """
    Use nslookup to lookup addresses
    :param name: Name of record to search
    :param rdtype: DNS record type
    :param timeout: server response timeout
    :param server: server to query
    :return: [] of records or False if error
    """
    cmd = f"nslookup -query={rdtype} {name}"

    if timeout is not None:
        cmd += f" -timeout={int(timeout)}"
    if server is not None:
        cmd += f" {server}"

    cmd = __salt__["cmd.run_all"](cmd, python_shell=False, output_loglevel="quiet")

    if cmd["retcode"] != 0:
        log.warning(
            "nslookup returned (%s): %s",
            cmd["retcode"],
            cmd["stdout"].splitlines()[-1].strip(string.whitespace + ";"),
        )
        return False

    lookup_res = iter(cmd["stdout"].splitlines())
    res = []
    try:
        line = next(lookup_res)
        if "unknown query type" in line:
            raise ValueError(f"Invalid DNS type {rdtype}")

        while True:
            if name in line:
                break
            line = next(lookup_res)

        while True:
            line = line.strip()
            if not line or line.startswith("*"):
                break
            elif rdtype != "CNAME" and "canonical name" in line:
                name = line.split()[-1][:-1]
                line = next(lookup_res)
                continue
            elif rdtype == "SOA":
                line = line.split("=")
            elif line.startswith("Name:"):
                line = next(lookup_res)
                line = line.split(":", 1)
            elif line.startswith(name):
                if "=" in line:
                    line = line.split("=", 1)
                else:
                    line = line.split(" ")

            res.append(_data_clean(line[-1]))
            line = next(lookup_res)

    except StopIteration:
        pass

    if rdtype == "SOA":
        return [" ".join(res[1:])]
    else:
        return res


def lookup(
    name,
    rdtype,
    method=None,
    servers=None,
    timeout=None,
    walk=False,
    walk_tld=False,
    secure=None,
):
    """
    Lookup DNS records and return their data

    :param name: name to lookup
    :param rdtype: DNS record type
    :param method: gai (getaddrinfo()), dnspython, dig, drill, host, nslookup or auto (default)
    :param servers: (list of) server(s) to try in-order
    :param timeout: query timeout or a valiant approximation of that
    :param walk: Walk the DNS upwards looking for the record type or name/recordtype if walk='name'.
    :param walk_tld: Include the final domain in the walk
    :param secure: return only DNSSEC secured responses
    :return: [] of record data
    """
    # opts = __opts__.get('dns', {})
    opts = {}
    method = method or opts.get("method", "auto")
    secure = secure or opts.get("secure", None)
    servers = servers or opts.get("servers", None)
    timeout = timeout or opts.get("timeout", False)

    rdtype = rdtype.upper()

    query_methods = (
        ("gai", _lookup_gai, not any((rdtype not in ("A", "AAAA"), servers, secure))),
        ("dnspython", _lookup_dnspython, HAS_DNSPYTHON),
        ("dig", _lookup_dig, HAS_DIG),
        ("drill", _lookup_drill, HAS_DRILL),
        ("host", _lookup_host, HAS_HOST and not secure),
        ("nslookup", _lookup_nslookup, HAS_NSLOOKUP and not secure),
    )

    try:
        if method == "auto":
            # The first one not to bork on the conditions becomes the function
            method, resolver = next(
                ((rname, rcb) for rname, rcb, rtest in query_methods if rtest)
            )
        else:
            # The first one not to bork on the conditions becomes the function. And the name must match.
            resolver = next(
                (
                    rcb
                    for rname, rcb, rtest in query_methods
                    if rname == method and rtest
                )
            )
    except StopIteration:
        log.error(
            "Unable to lookup %s/%s: Resolver method %s invalid, unsupported "
            "or unable to perform query",
            method,
            rdtype,
            name,
        )
        return False

    res_kwargs = {
        "rdtype": rdtype,
    }

    if servers:
        if not isinstance(servers, (list, tuple)):
            servers = [servers]
        if method in ("dnspython", "dig", "drill"):
            res_kwargs["servers"] = servers
        else:
            if timeout:
                timeout /= len(servers)

            # Inject a wrapper for multi-server behaviour
            def _multi_srvr(resolv_func):
                @functools.wraps(resolv_func)
                def _wrapper(**res_kwargs):
                    for server in servers:
                        s_res = resolv_func(server=server, **res_kwargs)
                        if s_res:
                            return s_res

                return _wrapper

            resolver = _multi_srvr(resolver)

    if not walk:
        name = [name]
    else:
        idx = 0
        if rdtype in ("SRV", "TLSA"):  # The only RRs I know that have 2 name components
            idx = name.find(".") + 1
        idx = name.find(".", idx) + 1
        domain = name[idx:]
        rname = name[0:idx]

        name = _tree(domain, walk_tld)
        if walk == "name":
            name = [rname + domain for domain in name]

        if timeout:
            timeout /= len(name)

    if secure:
        res_kwargs["secure"] = secure
    if timeout:
        res_kwargs["timeout"] = timeout

    for rname in name:
        res = resolver(name=rname, **res_kwargs)
        if res:
            return res

    return res


def query(
    name,
    rdtype,
    method=None,
    servers=None,
    timeout=None,
    walk=False,
    walk_tld=False,
    secure=None,
):
    """
    Query DNS for information.
    Where `lookup()` returns record data, `query()` tries to interpret the data and return its results

    :param name: name to lookup
    :param rdtype: DNS record type
    :param method: gai (getaddrinfo()), pydns, dig, drill, host, nslookup or auto (default)
    :param servers: (list of) server(s) to try in-order
    :param timeout: query timeout or a valiant approximation of that
    :param secure: return only DNSSEC secured response
    :param walk: Walk the DNS upwards looking for the record type or name/recordtype if walk='name'.
    :param walk_tld: Include the top-level domain in the walk
    :return: [] of records
    """
    rdtype = rdtype.upper()
    qargs = {
        "method": method,
        "servers": servers,
        "timeout": timeout,
        "walk": walk,
        "walk_tld": walk_tld,
        "secure": secure,
    }

    if rdtype == "PTR" and not name.endswith("arpa"):
        name = ptr_name(name)

    if rdtype == "SPF":
        # 'SPF' has become a regular 'TXT' again
        qres = [
            answer
            for answer in lookup(name, "TXT", **qargs)
            if answer.startswith("v=spf")
        ]
        if not qres:
            qres = lookup(name, rdtype, **qargs)
    else:
        qres = lookup(name, rdtype, **qargs)

    rec_map = {
        "A": a_rec,
        "AAAA": aaaa_rec,
        "CAA": caa_rec,
        "MX": mx_rec,
        "SOA": soa_rec,
        "SPF": spf_rec,
        "SRV": srv_rec,
        "SSHFP": sshfp_rec,
        "TLSA": tlsa_rec,
    }

    if not qres or rdtype not in rec_map:
        return qres
    elif rdtype in ("A", "AAAA", "SSHFP", "TLSA"):
        res = [rec_map[rdtype](res) for res in qres]
    elif rdtype in ("SOA", "SPF"):
        res = rec_map[rdtype](qres[0])
    else:
        res = rec_map[rdtype](qres)

    return res


def host(name, ip4=True, ip6=True, **kwargs):
    """
    Return a list of addresses for name

    ip6:
        Return IPv6 addresses
    ip4:
        Return IPv4 addresses

    the rest is passed on to lookup()
    """
    res = {}
    if ip6:
        ip6 = lookup(name, "AAAA", **kwargs)
        if ip6:
            res["ip6"] = ip6
    if ip4:
        ip4 = lookup(name, "A", **kwargs)
        if ip4:
            res["ip4"] = ip4

    return res


def a_rec(rdata):
    """
    Validate and parse DNS record data for an A record
    :param rdata: DNS record data
    :return: { 'address': ip }
    """
    rschema = OrderedDict((("address", ipaddress.IPv4Address),))
    return _data2rec(rschema, rdata)


def aaaa_rec(rdata):
    """
    Validate and parse DNS record data for an AAAA record
    :param rdata: DNS record data
    :return: { 'address': ip }
    """
    rschema = OrderedDict((("address", ipaddress.IPv6Address),))
    return _data2rec(rschema, rdata)


def caa_rec(rdatas):
    """
    Validate and parse DNS record data for a CAA record
    :param rdata: DNS record data
    :return: dict w/fields
    """
    rschema = OrderedDict(
        (
            ("flags", lambda flag: ["critical"] if int(flag) > 0 else []),
            ("tag", RFC.CAA_TAGS),
            ("value", lambda val: val.strip("',\"")),
        )
    )

    res = _data2rec_group(rschema, rdatas, "tag")

    for tag in ("issue", "issuewild"):
        tag_res = res.get(tag, False)
        if not tag_res:
            continue
        for idx, val in enumerate(tag_res):
            if ";" not in val:
                continue
            val, params = val.split(";", 1)
            params = dict(param.split("=") for param in shlex.split(params))
            tag_res[idx] = {val: params}

    return res


def mx_data(target, preference=10):
    """
    Generate MX record data
    :param target: server
    :param preference: preference number
    :return: DNS record data
    """
    return _rec2data(int(preference), target)


def mx_rec(rdatas):
    """
    Validate and parse DNS record data for MX record(s)
    :param rdata: DNS record data
    :return: dict w/fields
    """
    rschema = OrderedDict(
        (
            ("preference", int),
            ("name", str),
        )
    )
    return _data2rec_group(rschema, rdatas, "preference")


def ptr_name(rdata):
    """
    Return PTR name of given IP
    :param rdata: IP address
    :return: PTR record name
    """
    try:
        return ipaddress.ip_address(rdata).reverse_pointer
    except ValueError:
        log.error("Unable to generate PTR record; %s is not a valid IP address", rdata)
        return False


def soa_rec(rdata):
    """
    Validate and parse DNS record data for SOA record(s)
    :param rdata: DNS record data
    :return: dict w/fields
    """
    rschema = OrderedDict(
        (
            ("mname", str),
            ("rname", str),
            ("serial", int),
            ("refresh", int),
            ("retry", int),
            ("expire", int),
            ("minimum", int),
        )
    )
    return _data2rec(rschema, rdata)


def spf_rec(rdata):
    """
    Validate and parse DNS record data for SPF record(s)
    :param rdata: DNS record data
    :return: dict w/fields
    """
    spf_fields = rdata.split(" ")
    if not spf_fields.pop(0).startswith("v=spf"):
        raise ValueError("Not an SPF record")

    res = OrderedDict()
    mods = set()
    for mech_spec in spf_fields:
        if mech_spec.startswith(("exp", "redirect")):
            # It's a modifier
            mod, val = mech_spec.split("=", 1)
            if mod in mods:
                raise KeyError(f"Modifier {mod} can only appear once")

            mods.add(mod)
            continue

            # TODO: Should be in something intelligent like an SPF_get
            # if mod == 'exp':
            #     res[mod] = lookup(val, 'TXT', **qargs)
            #     continue
            # elif mod == 'redirect':
            #     return query(val, 'SPF', **qargs)

        mech = {}
        if mech_spec[0] in ("+", "-", "~", "?"):
            mech["qualifier"] = mech_spec[0]
            mech_spec = mech_spec[1:]

        if ":" in mech_spec:
            mech_spec, val = mech_spec.split(":", 1)
        elif "/" in mech_spec:
            idx = mech_spec.find("/")
            mech_spec = mech_spec[0:idx]
            val = mech_spec[idx:]
        else:
            val = None

        res[mech_spec] = mech
        if not val:
            continue
        elif mech_spec in ("ip4", "ip6"):
            val = ipaddress.ip_interface(val)
            assert val.version == int(mech_spec[-1])

        mech["value"] = val

    return res


def srv_data(target, port, prio=10, weight=10):
    """
    Generate SRV record data
    :param target:
    :param port:
    :param prio:
    :param weight:
    :return:
    """
    return _rec2data(prio, weight, port, target)


def srv_name(svc, proto="tcp", domain=None):
    """
    Generate SRV record name
    :param svc: ldap, 389 etc
    :param proto: tcp, udp, sctp etc.
    :param domain: name to append
    :return:
    """
    proto = RFC.validate(proto, RFC.SRV_PROTO)
    if isinstance(svc, int) or svc.isdigit():
        svc = _to_port(svc)

    if domain:
        domain = "." + domain
    return f"_{svc}._{proto}{domain}"


def srv_rec(rdatas):
    """
    Validate and parse DNS record data for SRV record(s)
    :param rdata: DNS record data
    :return: dict w/fields
    """
    rschema = OrderedDict(
        (
            ("prio", int),
            ("weight", int),
            ("port", _to_port),
            ("name", str),
        )
    )
    return _data2rec_group(rschema, rdatas, "prio")


def sshfp_data(key_t, hash_t, pub):
    """
    Generate an SSHFP record
    :param key_t: rsa/dsa/ecdsa/ed25519
    :param hash_t: sha1/sha256
    :param pub: the SSH public key
    """
    key_t = RFC.validate(key_t, RFC.SSHFP_ALGO, "in")
    hash_t = RFC.validate(hash_t, RFC.SSHFP_HASH)

    hasher = hashlib.new(hash_t)
    hasher.update(base64.b64decode(pub))
    ssh_fp = hasher.hexdigest()

    return _rec2data(key_t, hash_t, ssh_fp)


def sshfp_rec(rdata):
    """
    Validate and parse DNS record data for TLSA record(s)
    :param rdata: DNS record data
    :return: dict w/fields
    """
    rschema = OrderedDict(
        (
            ("algorithm", RFC.SSHFP_ALGO),
            ("fp_hash", RFC.SSHFP_HASH),
            (
                "fingerprint",
                lambda val: val.lower(),
            ),  # resolvers are inconsistent on this one
        )
    )

    return _data2rec(rschema, rdata)


def tlsa_data(pub, usage, selector, matching):
    """
    Generate a TLSA rec
    :param pub: Pub key in PEM format
    :param usage:
    :param selector:
    :param matching:
    :return: TLSA data portion
    """
    usage = RFC.validate(usage, RFC.TLSA_USAGE)
    selector = RFC.validate(selector, RFC.TLSA_SELECT)
    matching = RFC.validate(matching, RFC.TLSA_MATCHING)

    pub = ssl.PEM_cert_to_DER_cert(pub.strip())
    if matching == 0:
        cert_fp = binascii.b2a_hex(pub)
    else:
        hasher = hashlib.new(RFC.TLSA_MATCHING[matching])
        hasher.update(pub)
        cert_fp = hasher.hexdigest()

    return _rec2data(usage, selector, matching, cert_fp)


def tlsa_rec(rdata):
    """
    Validate and parse DNS record data for TLSA record(s)
    :param rdata: DNS record data
    :return: dict w/fields
    """
    rschema = OrderedDict(
        (
            ("usage", RFC.TLSA_USAGE),
            ("selector", RFC.TLSA_SELECT),
            ("matching", RFC.TLSA_MATCHING),
            ("pub", str),
        )
    )

    return _data2rec(rschema, rdata)


def service(svc, proto="tcp", domain=None, walk=False, secure=None):
    """
    Find an SRV service in a domain or its parents
    :param svc: service to find (ldap, 389, etc)
    :param proto: protocol the service talks (tcp, udp, etc)
    :param domain: domain to start search in
    :param walk: walk the parents if domain doesn't provide the service
    :param secure: only return DNSSEC-validated results
    :return: [
        [ prio1server1, prio1server2 ],
        [ prio2server1, prio2server2 ],
    ] (the servers will already be weighted according to the SRV rules)
    """
    qres = query(srv_name(svc, proto, domain), "SRV", walk=walk, secure=secure)
    if not qres:
        return False

    res = []
    for _, recs in qres.items():
        res.append(_weighted_order(recs))

    return res


def services(services_file="/etc/services"):
    """
    Parse through system-known services
    :return: {
        'svc': [
          {  'port': port
             'proto': proto,
             'desc': comment
          },
        ],
    }
    """
    res = {}
    with salt.utils.files.fopen(services_file, "r") as svc_defs:
        for svc_def in svc_defs.readlines():
            svc_def = salt.utils.stringutils.to_unicode(svc_def.strip())
            if not svc_def or svc_def.startswith("#"):
                continue
            elif "#" in svc_def:
                svc_def, comment = svc_def.split("#", 1)
                comment = comment.strip()
            else:
                comment = None
            svc_def = svc_def.split()

            port, proto = svc_def.pop(1).split("/")
            port = int(port)

            for name in svc_def:
                svc_res = res.get(name, {})
                pp_res = svc_res.get(port, False)
                if not pp_res:
                    svc = {
                        "port": port,
                        "proto": proto,
                    }
                    if comment:
                        svc["desc"] = comment
                    svc_res[port] = svc
                else:
                    curr_proto = pp_res["proto"]
                    if isinstance(curr_proto, (list, tuple)):
                        curr_proto.append(proto)
                    else:
                        pp_res["proto"] = [curr_proto, proto]

                    curr_desc = pp_res.get("desc", False)
                    if comment:
                        if not curr_desc:
                            pp_res["desc"] = comment
                        elif comment != curr_desc:
                            pp_res["desc"] = f"{curr_desc}, {comment}"
                res[name] = svc_res

    for svc, data in res.items():
        if len(data) == 1:
            res[svc] = data.values().pop()
            continue
        else:
            res[svc] = list(data.values())

    return res


def parse_resolv(src="/etc/resolv.conf"):
    """
    Parse a resolver configuration file (traditionally /etc/resolv.conf)
    """

    nameservers = []
    ip4_nameservers = []
    ip6_nameservers = []
    search = []
    sortlist = []
    domain = ""
    options = []

    try:
        with salt.utils.files.fopen(src) as src_file:
            # pylint: disable=too-many-nested-blocks
            for line in src_file:
                line = salt.utils.stringutils.to_unicode(line).strip().split()

                try:
                    (directive, arg) = (line[0].lower(), line[1:])
                    # Drop everything after # or ; (comments)
                    arg = list(
                        itertools.takewhile(lambda x: x[0] not in ("#", ";"), arg)
                    )
                    if directive == "nameserver":
                        addr = arg[0]
                        try:
                            ip_addr = ipaddress.ip_address(addr)
                            version = ip_addr.version
                            ip_addr = str(ip_addr)
                            if ip_addr not in nameservers:
                                nameservers.append(ip_addr)
                            if version == 4 and ip_addr not in ip4_nameservers:
                                ip4_nameservers.append(ip_addr)
                            elif version == 6 and ip_addr not in ip6_nameservers:
                                ip6_nameservers.append(ip_addr)
                        except ValueError as exc:
                            log.error("%s: %s", src, exc)
                    elif directive == "domain":
                        domain = arg[0]
                    elif directive == "search":
                        search = arg
                    elif directive == "sortlist":
                        # A sortlist is specified by IP address netmask pairs.
                        # The netmask is optional and defaults to the natural
                        # netmask of the net. The IP address and optional
                        # network pairs are separated by slashes.
                        for ip_raw in arg:
                            try:
                                ip_net = ipaddress.ip_network(ip_raw)
                            except ValueError as exc:
                                log.error("%s: %s", src, exc)
                            else:
                                if "/" not in ip_raw:
                                    # No netmask has been provided, guess
                                    # the "natural" one
                                    if ip_net.version == 4:
                                        ip_addr = str(ip_net.network_address)
                                        # pylint: disable=protected-access
                                        mask = salt.utils.network.natural_ipv4_netmask(
                                            ip_addr
                                        )
                                        ip_net = ipaddress.ip_network(
                                            f"{ip_addr}{mask}", strict=False
                                        )
                                    if ip_net.version == 6:
                                        # TODO
                                        pass

                                if ip_net not in sortlist:
                                    sortlist.append(ip_net)
                    elif directive == "options":
                        # Options allows certain internal resolver variables to
                        # be modified.
                        if arg[0] not in options:
                            options.append(arg[0])
                except IndexError:
                    continue

        if domain and search:
            # The domain and search keywords are mutually exclusive.  If more
            # than one instance of these keywords is present, the last instance
            # will override.
            log.debug("%s: The domain and search keywords are mutually exclusive.", src)

        return {
            "nameservers": nameservers,
            "ip4_nameservers": ip4_nameservers,
            "ip6_nameservers": ip6_nameservers,
            "sortlist": [ip.with_netmask for ip in sortlist],
            "domain": domain,
            "search": search,
            "options": options,
        }
    except OSError:
        return {}



"""
Integration tests for the rabbitmq_user states
"""

import logging

import pytest

import salt.modules.rabbitmq as rabbitmq
import salt.states.rabbitmq_vhost as rabbitmq_vhost

log = logging.getLogger(__name__)

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing(
        "docker", "dockerd", reason="Docker not installed"
    ),
]


@pytest.fixture
def configure_loader_modules(docker_cmd_run_all_wrapper):
    return {
        rabbitmq_vhost: {
            "__salt__": {
                "rabbitmq.vhost_exists": rabbitmq.vhost_exists,
                "rabbitmq.add_vhost": rabbitmq.add_vhost,
                "rabbitmq.delete_vhost": rabbitmq.delete_vhost,
            },
            "__opts__": {"test": False},
            "_utils__": {},
        },
        rabbitmq: {
            "__salt__": {"cmd.run_all": docker_cmd_run_all_wrapper},
            "__opts__": {},
            "_utils__": {},
        },
    }


def test_present_absent(docker_cmd_run_all_wrapper):
    """
    Test rabbitmq_vhost.present
    """

    # Clear the user
    ret = rabbitmq_vhost.present("vhost")
    expected = {
        "name": "vhost",
        "result": True,
        "comment": 'Adding vhost "vhost" ...\n',
        "changes": {"old": "", "new": "vhost"},
    }
    assert ret == expected

    # Delete the user
    ret = rabbitmq_vhost.absent("vhost")
    expected = {
        "name": "vhost",
        "result": True,
        "comment": 'Deleting vhost "vhost" ...\n',
        "changes": {"old": "vhost", "new": ""},
    }

    assert ret == expected


def test_absent(docker_cmd_run_all_wrapper):
    """
    Test rabbitmq_vhost.present
    """

    # Delete the user
    ret = rabbitmq_vhost.absent("vhost")
    expected = {
        "name": "vhost",
        "result": True,
        "comment": "Virtual Host 'vhost' is not present.",
        "changes": {},
    }
    assert ret == expected



import salt.exceptions


class VaultException(salt.exceptions.SaltException):
    """
    Base class for exceptions raised by this module
    """


class VaultLeaseExpired(VaultException):
    """
    Raised when a cached lease is reported to be expired locally.
    """


class VaultAuthExpired(VaultException):
    """
    Raised when cached authentication data is reported to be outdated locally.
    """


class VaultConfigExpired(VaultException):
    """
    Raised when secret authentication data queried from the master reports
    a different server configuration than locally cached or an explicit
    cache TTL set in the configuration has been reached.
    """


class VaultUnwrapException(VaultException):
    """
    Raised when an expected creation path for a wrapping token differs
    from the reported one.
    This has to be taken seriously as it indicates tampering.
    """

    def __init__(self, expected, actual, url, namespace, verify, *args, **kwargs):
        msg = (
            "Wrapped response was not created from expected Vault path: "
            f"`{actual}` is not matched by any of `{expected}`.\n"
            "This indicates tampering with the wrapping token by a third party "
            "and should be taken very seriously! If you changed some authentication-"
            "specific configuration on the master recently, especially minion "
            "approle mount, you should consider if this error was caused by outdated "
            "cached data on this minion instead."
        )
        super().__init__(msg, *args, **kwargs)
        self.event_data = {
            "expected": expected,
            "actual": actual,
            "url": url,
            "namespace": namespace,
            "verify": verify,
        }


# https://www.vaultproject.io/api-docs#http-status-codes
class VaultInvocationError(VaultException):
    """
    HTTP 400 and InvalidArgumentException for this module
    """


class VaultPermissionDeniedError(VaultException):
    """
    HTTP 403
    """


class VaultNotFoundError(VaultException):
    """
    HTTP 404
    In some cases, this is also raised when the client does not have
    the correct permissions for the requested endpoint.
    """


class VaultUnsupportedOperationError(VaultException):
    """
    HTTP 405
    """


class VaultPreconditionFailedError(VaultException):
    """
    HTTP 412
    """


class VaultServerError(VaultException):
    """
    HTTP 500
    HTTP 502
    """


class VaultUnavailableError(VaultException):
    """
    HTTP 503
    Indicates maintenance or sealed status.
    """



"""
    :synopsis: Unit Tests for 'module.aptkernelpkg'
    :platform: Linux
    :maturity: develop
    .. versionadded:: 2018.3.0
"""

# pylint: disable=invalid-name,no-member

import re

import pytest

try:
    # Import Salt Testing Libs
    import salt.modules.kernelpkg_linux_apt as kernelpkg
    from salt.exceptions import CommandExecutionError

    # Import Salt Libs
    from tests.support.kernelpkg import KernelPkgTestCase
    from tests.support.mixins import LoaderModuleMockMixin
    from tests.support.mock import MagicMock, patch
    from tests.support.unit import TestCase

    HAS_MODULES = True
except ImportError:
    HAS_MODULES = False


@pytest.mark.skipif(not HAS_MODULES, reason="Salt modules could not be loaded")
class AptKernelPkgTestCase(KernelPkgTestCase, TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.kernelpkg_linux_apt
    """

    _kernelpkg = kernelpkg
    KERNEL_LIST = ["4.4.0-70-generic", "4.4.0-71-generic", "4.5.1-14-generic"]
    PACKAGE_DICT = {}

    @classmethod
    def setUpClass(cls):
        version = re.match(r"^(\d+\.\d+\.\d+)-(\d+)", cls.KERNEL_LIST[-1])
        cls.LATEST = f"{version.group(1)}.{version.group(2)}"

        for kernel in cls.KERNEL_LIST:
            pkg = "{}-{}".format(
                kernelpkg._package_prefix(), kernel
            )  # pylint: disable=protected-access
            cls.PACKAGE_DICT[pkg] = pkg

    def setup_loader_modules(self):
        return {
            kernelpkg: {
                "__grains__": {"kernelrelease": self.KERNEL_LIST[0]},
                "__salt__": {
                    "pkg.install": MagicMock(return_value={}),
                    "pkg.latest_version": MagicMock(return_value=self.LATEST),
                    "pkg.list_pkgs": MagicMock(return_value=self.PACKAGE_DICT),
                    "pkg.purge": MagicMock(return_value=None),
                    "system.reboot": MagicMock(return_value=None),
                },
            }
        }

    def test_list_installed(self):
        """
        Test - Return return the latest installed kernel version
        """
        PACKAGE_LIST = [
            f"{kernelpkg._package_prefix()}-{kernel}" for kernel in self.KERNEL_LIST
        ]  # pylint: disable=protected-access

        mock = MagicMock(return_value=PACKAGE_LIST)
        with patch.dict(self._kernelpkg.__salt__, {"pkg.list_pkgs": mock}):
            self.assertListEqual(self._kernelpkg.list_installed(), self.KERNEL_LIST)

    def test_list_installed_none(self):
        """
        Test - Return return the latest installed kernel version
        """
        mock = MagicMock(return_value=None)
        with patch.dict(self._kernelpkg.__salt__, {"pkg.list_pkgs": mock}):
            self.assertListEqual(self._kernelpkg.list_installed(), [])

    def test_remove_success(self):
        """
        Test - remove kernel package
        """
        with patch.object(self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]):
            with patch.object(
                self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
            ):
                result = self._kernelpkg.remove(release=self.KERNEL_LIST[0])
                self.assertIn("removed", result)
                target = "{}-{}".format(
                    self._kernelpkg._package_prefix(), self.KERNEL_LIST[0]
                )  # pylint: disable=protected-access
                self.assertListEqual(result["removed"], [target])

    def test_remove_error(self):
        """
        Test - remove kernel package
        """
        mock = MagicMock(side_effect=CommandExecutionError())
        with patch.dict(self._kernelpkg.__salt__, {"pkg.purge": mock}):
            with patch.object(
                self._kernelpkg, "active", return_value=self.KERNEL_LIST[-1]
            ):
                with patch.object(
                    self._kernelpkg, "list_installed", return_value=self.KERNEL_LIST
                ):
                    self.assertRaises(
                        CommandExecutionError,
                        self._kernelpkg.remove,
                        release=self.KERNEL_LIST[0],
                    )



import subprocess


def test_help(install_salt):
    """
    Test --help works for all salt cmds
    """
    for cmd in install_salt.binary_paths.values():
        cmd = [str(x) for x in cmd]

        if len(cmd) > 1 and "shell" in cmd[1]:
            # Singlebin build, unable to get the version
            continue

        if "python" in cmd[0] and len(cmd) == 1:
            ret = install_salt.proc.run(
                *cmd, "--version", stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            assert "Python" in ret.stdout
        else:
            ret = install_salt.proc.run(
                *cmd, "--help", stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            assert "Usage" in ret.stdout
            assert ret.returncode == 0



"""
Create an XDG function to get the config dir
"""

import os


def xdg_config_dir():
    """
    Check xdg locations for config files
    """
    xdg_config = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
    xdg_config_directory = os.path.join(xdg_config, "salt")
    return xdg_config_directory



import pytest

from . import normalize_ret

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_requisites_onfail_any(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    a:
      cmd.run:
        - name: exit 0

    b:
      cmd.run:
        - name: exit 1

    c:
      cmd.run:
        - name: exit 0

    d:
      cmd.run:
        - name: echo itworked
        - onfail_any:
          - cmd: a
          - cmd: b
          - cmd: c

    e:
      cmd.run:
        - name: exit 0

    f:
      cmd.run:
        - name: exit 0

    g:
      cmd.run:
        - name: exit 0

    h:
      cmd.run:
        - name: echo itworked
        - onfail_any:
          - cmd: e
          - cmd: f
          - cmd: g
    """
    expected_result = {
        "cmd_|-a_|-exit 0_|-run": {
            "__run_num__": 0,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-b_|-exit 1_|-run": {
            "__run_num__": 1,
            "changes": True,
            "comment": 'Command "exit 1" run',
            "result": False,
        },
        "cmd_|-c_|-exit 0_|-run": {
            "__run_num__": 2,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-d_|-echo itworked_|-run": {
            "__run_num__": 3,
            "changes": True,
            "comment": 'Command "echo itworked" run',
            "result": True,
        },
        "cmd_|-e_|-exit 0_|-run": {
            "__run_num__": 4,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-f_|-exit 0_|-run": {
            "__run_num__": 5,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-g_|-exit 0_|-run": {
            "__run_num__": 6,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-h_|-echo itworked_|-run": {
            "__run_num__": 7,
            "changes": False,
            "comment": "State was not run because onfail req did not change",
            "result": True,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_onfail_all(state, state_tree):
    """
    Call sls file containing several onfail-all

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    a:
      cmd.run:
        - name: exit 0

    b:
      cmd.run:
        - name: exit 0

    c:
      cmd.run:
        - name: exit 0

    d:
      cmd.run:
        - name: exit 1

    e:
      cmd.run:
        - name: exit 1

    f:
      cmd.run:
        - name: exit 1

    reqs not met:
      cmd.run:
        - name: echo itdidntonfail
        - onfail_all:
          - cmd: a
          - cmd: e

    reqs also not met:
      cmd.run:
        - name: echo italsodidnonfail
        - onfail_all:
          - cmd: a
          - cmd: b
          - cmd: c

    reqs met:
      cmd.run:
        - name: echo itonfailed
        - onfail_all:
          - cmd: d
          - cmd: e
          - cmd: f

    reqs also met:
      cmd.run:
        - name: echo itonfailed
        - onfail_all:
          - cmd: d
        - require:
          - cmd: a
    """
    expected_result = {
        "cmd_|-a_|-exit 0_|-run": {
            "__run_num__": 0,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-b_|-exit 0_|-run": {
            "__run_num__": 1,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-c_|-exit 0_|-run": {
            "__run_num__": 2,
            "changes": True,
            "comment": 'Command "exit 0" run',
            "result": True,
        },
        "cmd_|-d_|-exit 1_|-run": {
            "__run_num__": 3,
            "changes": True,
            "comment": 'Command "exit 1" run',
            "result": False,
        },
        "cmd_|-e_|-exit 1_|-run": {
            "__run_num__": 4,
            "changes": True,
            "comment": 'Command "exit 1" run',
            "result": False,
        },
        "cmd_|-f_|-exit 1_|-run": {
            "__run_num__": 5,
            "changes": True,
            "comment": 'Command "exit 1" run',
            "result": False,
        },
        "cmd_|-reqs also met_|-echo itonfailed_|-run": {
            "__run_num__": 9,
            "changes": True,
            "comment": 'Command "echo itonfailed" run',
            "result": True,
        },
        "cmd_|-reqs also not met_|-echo italsodidnonfail_|-run": {
            "__run_num__": 7,
            "changes": False,
            "comment": "State was not run because onfail req did not change",
            "result": True,
        },
        "cmd_|-reqs met_|-echo itonfailed_|-run": {
            "__run_num__": 8,
            "changes": True,
            "comment": 'Command "echo itonfailed" run',
            "result": True,
        },
        "cmd_|-reqs not met_|-echo itdidntonfail_|-run": {
            "__run_num__": 6,
            "changes": False,
            "comment": "State was not run because onfail req did not change",
            "result": True,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_onfail_requisite(state, state_tree):
    """
    Tests a simple state using the onfail requisite
    """
    sls_contents = """
    failing_state:
      cmd.run:
        - name: asdf

    non_failing_state:
      cmd.run:
        - name: echo "Non-failing state"

    test_failing_state:
      cmd.run:
        - name: echo "Success!"
        - onfail:
          - cmd: failing_state

    test_non_failing_state:
      cmd.run:
        - name: echo "Should not run"
        - onfail:
          - cmd: non_failing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_failing_state_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )

        assert (
            ret['cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'].comment
            == "State was not run because onfail req did not change"
        )


def test_multiple_onfail_requisite(state, state_tree):
    """
    test to ensure state is run even if only one
    of the onfails fails. This is a test for the issue:
    https://github.com/saltstack/salt/issues/22370
    """
    sls_contents = """
    a:
      cmd.run:
        - name: exit 0

    b:
      cmd.run:
        - name: exit 1

    c:
      cmd.run:
        - name: echo itworked
        - onfail:
          - cmd: a
          - cmd: b
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret["cmd_|-c_|-echo itworked_|-run"].changes["retcode"] == 0
        assert ret["cmd_|-c_|-echo itworked_|-run"].changes["stdout"] == "itworked"


def test_onfail_in_requisite(state, state_tree):
    """
    Tests a simple state using the onfail_in requisite
    """
    sls_contents = """
    failing_state:
      cmd.run:
        - name: asdf
        - onfail_in:
          - cmd: test_failing_state

    non_failing_state:
      cmd.run:
        - name: echo "Non-failing state"
        - onfail_in:
          - cmd: test_non_failing_state

    test_failing_state:
      cmd.run:
        - name: echo "Success!"

    test_non_failing_state:
      cmd.run:
        - name: echo "Should not run"
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_failing_state_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )
        assert (
            ret['cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'].comment
            == "State was not run because onfail req did not change"
        )


def test_onfail_requisite_no_state_module(state, state_tree):
    """
    Tests a simple state using the onfail requisite
    """
    sls_contents = """
    failing_state:
      cmd.run:
        - name: asdf

    non_failing_state:
      cmd.run:
        - name: echo "Non-failing state"

    test_failing_state:
      cmd.run:
        - name: echo "Success!"
        - onfail:
          - failing_state

    test_non_failing_state:
      cmd.run:
        - name: echo "Should not run"
        - onfail:
          - non_failing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret['cmd_|-test_failing_state_|-echo "Success!"_|-run'].comment
            == 'Command "echo "Success!"" run'
        )
        assert (
            ret['cmd_|-test_non_failing_state_|-echo "Should not run"_|-run'].comment
            == "State was not run because onfail req did not change"
        )


def test_onfail_requisite_with_duration(state, state_tree):
    """
    Tests a simple state using the onfail requisite
    """
    sls_contents = """
    failing_state:
      cmd.run:
        - name: asdf

    non_failing_state:
      cmd.run:
        - name: echo "Non-failing state"

    test_failing_state:
      cmd.run:
        - name: echo "Success!"
        - onfail:
          - cmd: failing_state

    test_non_failing_state:
      cmd.run:
        - name: echo "Should not run"
        - onfail:
          - cmd: non_failing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            "duration"
            in ret['cmd_|-test_non_failing_state_|-echo "Should not run"_|-run']
        )


def test_multiple_onfail_requisite_with_required(state, state_tree):
    """
    test to ensure multiple states are run
    when specified as onfails for a single state.
    This is a test for the issue:
    https://github.com/saltstack/salt/issues/46552
    """
    sls_contents = """
    a:
      cmd.run:
        - name: exit 1

    pass:
      cmd.run:
        - name: exit 0

    b:
      cmd.run:
        - name: echo b
        - onfail:
          - cmd: a

    c:
      cmd.run:
        - name: echo c
        - onfail:
          - cmd: a
        - require:
          - cmd: b

    d:
      cmd.run:
        - name: echo d
        - onfail:
          - cmd: a
        - require:
          - cmd: c

    e:
      cmd.run:
        - name: echo e
        - onfail:
          - cmd: pass
        - require:
          - cmd: c

    f:
      cmd.run:
        - name: echo f
        - onfail:
          - cmd: pass
        - onchanges:
          - cmd: b
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")

        assert ret["cmd_|-b_|-echo b_|-run"].changes["retcode"] == 0
        assert ret["cmd_|-c_|-echo c_|-run"].changes["retcode"] == 0
        assert ret["cmd_|-d_|-echo d_|-run"].changes["retcode"] == 0
        assert ret["cmd_|-b_|-echo b_|-run"].changes["stdout"] == "b"
        assert ret["cmd_|-c_|-echo c_|-run"].changes["stdout"] == "c"
        assert ret["cmd_|-d_|-echo d_|-run"].changes["stdout"] == "d"
        assert (
            ret["cmd_|-e_|-echo e_|-run"].comment
            == "State was not run because onfail req did not change"
        )
        assert (
            ret["cmd_|-f_|-echo f_|-run"].comment
            == "State was not run because onfail req did not change"
        )


def test_multiple_onfail_requisite_with_required_no_run(state, state_tree):
    """
    test to ensure multiple states are not run
    when specified as onfails for a single state
    which fails.
    This is a test for the issue:
    https://github.com/saltstack/salt/issues/46552
    """
    sls_contents = """
    a:
      cmd.run:
        - name: exit 0

    b:
      cmd.run:
        - name: echo b
        - onfail:
          - cmd: a

    c:
      cmd.run:
        - name: echo c
        - onfail:
          - cmd: a
        - require:
          - cmd: b

    d:
      cmd.run:
        - name: echo d
        - onfail:
          - cmd: a
        - require:
          - cmd: c
    """
    expected = "State was not run because onfail req did not change"
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret["cmd_|-b_|-echo b_|-run"].comment == expected
        assert ret["cmd_|-c_|-echo c_|-run"].comment == expected
        assert ret["cmd_|-d_|-echo d_|-run"].comment == expected



"""
Manage accounts in Samba's passdb using pdbedit

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       pdbedit
:platform:      posix

.. versionadded:: 2017.7.0

.. code-block:: yaml

    wash:
      pdbedit.absent

    kaylee:
      pdbedit.managed:
        - password: A70C708517B5DD0EDB67714FE25336EB
        - password_hashed: True
        - drive: 'X:'
        - homedir: '\\\\serenity\\mechanic\\profile'
"""

import logging

import salt.utils.data

log = logging.getLogger(__name__)

# Define the state's virtual name
__virtualname__ = "pdbedit"


def __virtual__():
    """
    Provides pdbedit when available
    """
    if "pdbedit.create" in __salt__:
        return True
    else:
        return (
            False,
            "{} state module can only be loaded when the pdbedit module is available".format(
                __virtualname__
            ),
        )


def absent(name):
    """
    Ensure user account is absent

    name : string
        username
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # remove if needed
    if name in __salt__["pdbedit.list"](False):
        res = __salt__["pdbedit.delete"](name)
        if res[name] in ["deleted"]:  # check if we need to update changes
            ret["changes"].update(res)
        elif res[name] not in ["absent"]:  # oops something went wrong
            ret["result"] = False
    else:
        ret["comment"] = f"account {name} is absent"

    return ret


def managed(name, **kwargs):
    """
    Manage user account

    login : string
        login name
    password : string
        password
    password_hashed : boolean
        set if password is a nt hash instead of plain text
    domain : string
        users domain
    profile : string
        profile path
    script : string
        logon script
    drive : string
        home drive
    homedir : string
        home directory
    fullname : string
        full name
    account_desc : string
        account description
    machine_sid : string
        specify the machines new primary group SID or rid
    user_sid : string
        specify the users new primary group SID or rid
    account_control : string
        specify user account control properties

        .. note::
            Only the following can be set:
            - N: No password required
            - D: Account disabled
            - H: Home directory required
            - L: Automatic Locking
            - X: Password does not expire
    reset_login_hours : boolean
        reset the users allowed logon hours
    reset_bad_password_count : boolean
        reset the stored bad login counter
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    # save state
    saved = __salt__["pdbedit.list"](hashes=True)
    saved = saved[name] if name in saved else {}

    # call pdbedit.modify
    kwargs["login"] = name
    res = __salt__["pdbedit.modify"](**kwargs)

    # calculate changes
    if res[name] in ["created"]:
        ret["changes"] = res
    elif res[name] in ["updated"]:
        ret["changes"][name] = salt.utils.data.compare_dicts(
            saved,
            __salt__["pdbedit.list"](hashes=True)[name],
        )
    elif res[name] not in ["unchanged"]:
        ret["result"] = False
        ret["comment"] = res[name]

    return ret


def present(name, **kwargs):
    """
    Alias for pdbedit.managed
    """
    return managed(name, **kwargs)



"""
    Unit tests for salt.acl.ClientACL
"""

import pytest

from salt import acl


@pytest.fixture
def configure_loader_modules():
    return {acl: {}}


@pytest.fixture
def blacklist():
    return {
        "users": ["joker", "penguin", "*bad_*", "blocked_.*", "^Homer$"],
        "modules": ["cmd.run", "test.fib", "rm-rf.*"],
    }


@pytest.fixture
def whitelist():
    return {
        "users": ["testuser", "saltuser"],
        "modules": ["test.ping", "grains.items"],
    }


def test_user_is_blacklisted(blacklist):
    """
    test user_is_blacklisted
    """
    client_acl = acl.PublisherACL(blacklist)

    assert client_acl.user_is_blacklisted("joker")
    assert client_acl.user_is_blacklisted("penguin")
    assert client_acl.user_is_blacklisted("bad_")
    assert client_acl.user_is_blacklisted("bad_user")
    assert client_acl.user_is_blacklisted("bad_*")
    assert client_acl.user_is_blacklisted("user_bad_")
    assert client_acl.user_is_blacklisted("blocked_")
    assert client_acl.user_is_blacklisted("blocked_user")
    assert client_acl.user_is_blacklisted("blocked_.*")
    assert client_acl.user_is_blacklisted("Homer")

    assert not client_acl.user_is_blacklisted("batman")
    assert not client_acl.user_is_blacklisted("robin")
    assert not client_acl.user_is_blacklisted("bad")
    assert not client_acl.user_is_blacklisted("blocked")
    assert not client_acl.user_is_blacklisted("NotHomer")
    assert not client_acl.user_is_blacklisted("HomerSimpson")


def test_cmd_is_blacklisted(blacklist):
    """
    test cmd_is_blacklisted
    """
    client_acl = acl.PublisherACL(blacklist)

    assert client_acl.cmd_is_blacklisted("cmd.run")
    assert client_acl.cmd_is_blacklisted("test.fib")
    assert client_acl.cmd_is_blacklisted("rm-rf.root")

    assert not client_acl.cmd_is_blacklisted("cmd.shell")
    assert not client_acl.cmd_is_blacklisted("test.versions")
    assert not client_acl.cmd_is_blacklisted("arm-rf.root")

    assert client_acl.cmd_is_blacklisted(["cmd.run", "state.sls"])
    assert not client_acl.cmd_is_blacklisted(["state.highstate", "state.sls"])


def test_user_is_whitelisted(whitelist):
    """
    test user_is_whitelisted
    """
    client_acl = acl.PublisherACL(whitelist)

    assert client_acl.user_is_whitelisted("testuser")
    assert client_acl.user_is_whitelisted("saltuser")
    assert not client_acl.user_is_whitelisted("three")
    assert not client_acl.user_is_whitelisted("hans")


def test_cmd_is_whitelisted(whitelist):
    """
    test cmd_is_whitelisted
    """
    client_acl = acl.PublisherACL(whitelist)

    assert client_acl.cmd_is_whitelisted("test.ping")
    assert client_acl.cmd_is_whitelisted("grains.items")
    assert not client_acl.cmd_is_whitelisted("cmd.run")
    assert not client_acl.cmd_is_whitelisted("test.version")



import os

import pytest


@pytest.fixture(scope="module")
def grains(modules):
    return modules.grains


def test_grains_items(grains):
    """
    Test running grains.items and assert
    certain information is included in
    the return
    """
    ret = grains.items()
    if os.environ.get("ONEDIR_TESTRUN", "0") == "0":
        assert ret["package"] == "pip"
    else:
        assert ret["package"] == "onedir"

    for key in ["num_cpus", "cpu_model", "os_family"]:
        assert key in ret.keys()



"""
Grains for NVMe Qualified Names (NQN).

.. versionadded:: 3000

To enable these grains set `nvme_grains: True` in the minion config.

.. code-block:: yaml

    nvme_grains: True
"""

import errno
import logging

import salt.utils.files
import salt.utils.path
import salt.utils.platform

__virtualname__ = "nvme"

# Get logging started
log = logging.getLogger(__name__)


def __virtual__():
    if __opts__.get("nvme_grains", False) is False:
        return False
    return __virtualname__


def nvme_nqn():
    """
    Return NVMe NQN
    """
    grains = {}
    grains["nvme_nqn"] = False
    if salt.utils.platform.is_linux():
        grains["nvme_nqn"] = _linux_nqn()
    return grains


def _linux_nqn():
    """
    Return NVMe NQN from a Linux host.
    """
    ret = []

    initiator = "/etc/nvme/hostnqn"
    try:
        with salt.utils.files.fopen(initiator, "r") as _nvme:
            for line in _nvme:
                line = line.strip()
                if line.startswith("nqn."):
                    ret.append(line)
    except OSError as ex:
        if ex.errno != errno.ENOENT:
            log.debug("Error while accessing '%s': %s", initiator, ex)

    return ret



"""
The ``file_tree`` external pillar allows values from all files in a directory
tree to be imported as Pillar data.

.. note::

    This is an external pillar and is subject to the :ref:`rules and
    constraints <external-pillars>` governing external pillars.

.. versionadded:: 2015.5.0

In this pillar, data is organized by either Minion ID or Nodegroup name.  To
setup pillar data for a specific Minion, place it in
``<root_dir>/hosts/<minion_id>``.  To setup pillar data for an entire
Nodegroup, place it in ``<root_dir>/nodegroups/<node_group>`` where
``<node_group>`` is the Nodegroup's name.

Example ``file_tree`` Pillar
============================

Master Configuration
--------------------

.. code-block:: yaml

    ext_pillar:
      - file_tree:
          root_dir: /srv/ext_pillar
          follow_dir_links: False
          keep_newline: True

The ``root_dir`` parameter is required and points to the directory where files
for each host are stored. The ``follow_dir_links`` parameter is optional and
defaults to False. If ``follow_dir_links`` is set to True, this external pillar
will follow symbolic links to other directories.

.. warning::
    Be careful when using ``follow_dir_links``, as a recursive symlink chain
    will result in unexpected results.

.. versionchanged:: 2018.3.0
    If ``root_dir`` is a relative path, it will be treated as relative to the
    :conf_master:`pillar_roots` of the environment specified by
    :conf_minion:`pillarenv`. If an environment specifies multiple
    roots, this module will search for files relative to all of them, in order,
    merging the results.

If ``keep_newline`` is set to ``True``, then the pillar values for files ending
in newlines will keep that newline. The default behavior is to remove the
end-of-file newline. ``keep_newline`` should be turned on if the pillar data is
intended to be used to deploy a file using ``contents_pillar`` with a
:py:func:`file.managed <salt.states.file.managed>` state.

.. versionchanged:: 2015.8.4
    The ``raw_data`` parameter has been renamed to ``keep_newline``. In earlier
    releases, ``raw_data`` must be used. Also, this parameter can now be a list
    of globs, allowing for more granular control over which pillar values keep
    their end-of-file newline. The globs match paths relative to the
    directories named for minion IDs and nodegroups underneath the ``root_dir``
    (see the layout examples in the below sections).

    .. code-block:: yaml

        ext_pillar:
          - file_tree:
              root_dir: /path/to/root/directory
              keep_newline:
                - files/testdir/*

.. note::
    In earlier releases, this documentation incorrectly stated that binary
    files would not affected by the ``keep_newline`` configuration.  However,
    this module does not actually distinguish between binary and text files.

.. versionchanged:: 2017.7.0
    Templating/rendering has been added. You can now specify a default render
    pipeline and a black- and whitelist of (dis)allowed renderers.

    ``template`` must be set to ``True`` for templating to happen.

    .. code-block:: yaml

        ext_pillar:
          - file_tree:
            root_dir: /path/to/root/directory
            render_default: jinja|yaml
            renderer_blacklist:
              - gpg
            renderer_whitelist:
              - jinja
              - yaml
            template: True

Assigning Pillar Data to Individual Hosts
-----------------------------------------

To configure pillar data for each host, this external pillar will recursively
iterate over ``root_dir``/hosts/``id`` (where ``id`` is a minion ID), and
compile pillar data with each subdirectory as a dictionary key and each file
as a value.

For example, the following ``root_dir`` tree:

.. code-block:: text

    ./hosts/
    ./hosts/test-host/
    ./hosts/test-host/files/
    ./hosts/test-host/files/testdir/
    ./hosts/test-host/files/testdir/file1.txt
    ./hosts/test-host/files/testdir/file2.txt
    ./hosts/test-host/files/another-testdir/
    ./hosts/test-host/files/another-testdir/symlink-to-file1.txt

will result in the following pillar tree for minion with ID ``test-host``:

.. code-block:: text

    test-host:
        ----------
        apache:
            ----------
            config.d:
                ----------
                00_important.conf:
                    <important_config important_setting="yes" />
                20_bob_extra.conf:
                    <bob_specific_cfg has_freeze_ray="yes" />
        corporate_app:
            ----------
            settings:
                ----------
                common_settings:
                    // This is the main settings file for the corporate
                    // internal web app
                    main_setting: probably
                bob_settings:
                    role: bob

.. note::

    The leaf data in the example shown is the contents of the pillar files.
"""

import fnmatch
import logging
import os

import salt.loader
import salt.template
import salt.utils.dictupdate
import salt.utils.files
import salt.utils.minions
import salt.utils.path
import salt.utils.stringio
import salt.utils.stringutils

# Set up logging
log = logging.getLogger(__name__)


def _on_walk_error(err):
    """
    Log salt.utils.path.os_walk() error.
    """
    log.error("%s: %s", err.filename, err.strerror)


def _check_newline(prefix, file_name, keep_newline):
    """
    Return a boolean stating whether or not a file's trailing newline should be
    removed. To figure this out, first check if keep_newline is a boolean and
    if so, return its opposite. Otherwise, iterate over keep_newline and check
    if any of the patterns match the file path. If a match is found, return
    False, otherwise return True.
    """
    if isinstance(keep_newline, bool):
        return not keep_newline
    full_path = os.path.join(prefix, file_name)
    for pattern in keep_newline:
        try:
            if fnmatch.fnmatch(full_path, pattern):
                return False
        except TypeError:
            if fnmatch.fnmatch(full_path, str(pattern)):
                return False
    return True


def _construct_pillar(
    top_dir,
    follow_dir_links,
    keep_newline=False,
    render_default=None,
    renderer_blacklist=None,
    renderer_whitelist=None,
    template=False,
):
    """
    Construct pillar from file tree.
    """
    pillar = {}
    renderers = salt.loader.render(__opts__, __salt__)

    norm_top_dir = os.path.normpath(top_dir)
    for dir_path, dir_names, file_names in salt.utils.path.os_walk(
        top_dir, topdown=True, onerror=_on_walk_error, followlinks=follow_dir_links
    ):
        # Find current path in pillar tree
        pillar_node = pillar
        norm_dir_path = os.path.normpath(dir_path)
        prefix = os.path.relpath(norm_dir_path, norm_top_dir)
        if norm_dir_path != norm_top_dir:
            path_parts = []
            head = prefix
            while head:
                head, tail = os.path.split(head)
                path_parts.insert(0, tail)
            while path_parts:
                pillar_node = pillar_node[path_parts.pop(0)]

        # Create dicts for subdirectories
        for dir_name in dir_names:
            pillar_node[dir_name] = {}

        # Add files
        for file_name in file_names:
            file_path = os.path.join(dir_path, file_name)
            if not os.path.isfile(file_path):
                log.error("file_tree: %s: not a regular file", file_path)
                continue

            contents = b""
            try:
                with salt.utils.files.fopen(file_path, "rb") as fhr:
                    buf = fhr.read(__opts__["file_buffer_size"])
                    while buf:
                        contents += buf
                        buf = fhr.read(__opts__["file_buffer_size"])
                    if contents.endswith(b"\n") and _check_newline(
                        prefix, file_name, keep_newline
                    ):
                        contents = contents[:-1]
            except OSError as exc:
                log.error("file_tree: Error reading %s: %s", file_path, exc.strerror)
            else:
                data = contents
                if template is True:
                    data = salt.template.compile_template_str(
                        template=salt.utils.stringutils.to_unicode(contents),
                        renderers=renderers,
                        default=render_default,
                        blacklist=renderer_blacklist,
                        whitelist=renderer_whitelist,
                    )
                if salt.utils.stringio.is_readable(data):
                    pillar_node[file_name] = data.getvalue()
                else:
                    pillar_node[file_name] = data

    return pillar


def ext_pillar(
    minion_id,
    pillar,
    root_dir=None,
    follow_dir_links=False,
    debug=False,
    keep_newline=False,
    render_default=None,
    renderer_blacklist=None,
    renderer_whitelist=None,
    template=False,
):
    """
    Compile pillar data from the given ``root_dir`` specific to Nodegroup names
    and Minion IDs.

    If a Minion's ID is not found at ``<root_dir>/host/<minion_id>`` or if it
    is not included in any Nodegroups named at
    ``<root_dir>/nodegroups/<node_group>``, no pillar data provided by this
    pillar module will be available for that Minion.

    .. versionchanged:: 2017.7.0
        Templating/rendering has been added. You can now specify a default
        render pipeline and a black- and whitelist of (dis)allowed renderers.

        ``template`` must be set to ``True`` for templating to happen.

        .. code-block:: yaml

            ext_pillar:
              - file_tree:
                root_dir: /path/to/root/directory
                render_default: jinja|yaml
                renderer_blacklist:
                  - gpg
                renderer_whitelist:
                  - jinja
                  - yaml
                template: True

    :param minion_id:
        The ID of the Minion whose pillar data is to be collected

    :param pillar:
        Unused by the ``file_tree`` pillar module

    :param root_dir:
        Filesystem directory used as the root for pillar data (e.g.
        ``/srv/ext_pillar``)

        .. versionchanged:: 2018.3.0
            If ``root_dir`` is a relative path, it will be treated as relative to the
            :conf_master:`pillar_roots` of the environment specified by
            :conf_minion:`pillarenv`. If an environment specifies multiple
            roots, this module will search for files relative to all of them, in order,
            merging the results.

    :param follow_dir_links:
        Follow symbolic links to directories while collecting pillar files.
        Defaults to ``False``.

        .. warning::

            Care should be exercised when enabling this option as it will
            follow links that point outside of ``root_dir``.

        .. warning::

            Symbolic links that lead to infinite recursion are not filtered.

    :param debug:
        Enable debug information at log level ``debug``.  Defaults to
        ``False``.  This option may be useful to help debug errors when setting
        up the ``file_tree`` pillar module.

    :param keep_newline:
        Preserve the end-of-file newline in files.  Defaults to ``False``.
        This option may either be a boolean or a list of file globs (as defined
        by the `Python fnmatch package
        <https://docs.python.org/library/fnmatch.html>`_) for which end-of-file
        newlines are to be kept.

        ``keep_newline`` should be turned on if the pillar data is intended to
        be used to deploy a file using ``contents_pillar`` with a
        :py:func:`file.managed <salt.states.file.managed>` state.

        .. versionchanged:: 2015.8.4
            The ``raw_data`` parameter has been renamed to ``keep_newline``. In
            earlier releases, ``raw_data`` must be used. Also, this parameter
            can now be a list of globs, allowing for more granular control over
            which pillar values keep their end-of-file newline. The globs match
            paths relative to the directories named for Minion IDs and
            Nodegroup namess underneath the ``root_dir``.

            .. code-block:: yaml

                ext_pillar:
                  - file_tree:
                      root_dir: /srv/ext_pillar
                      keep_newline:
                        - apache/config.d/*
                        - corporate_app/settings/*

        .. note::
            In earlier releases, this documentation incorrectly stated that
            binary files would not affected by the ``keep_newline``.  However,
            this module does not actually distinguish between binary and text
            files.


    :param render_default:
        Override Salt's :conf_master:`default global renderer <renderer>` for
        the ``file_tree`` pillar.

        .. code-block:: yaml

            render_default: jinja

    :param renderer_blacklist:
        Disallow renderers for pillar files.

        .. code-block:: yaml

            renderer_blacklist:
              - json

    :param renderer_whitelist:
        Allow renderers for pillar files.

        .. code-block:: yaml

            renderer_whitelist:
              - yaml
              - jinja

    :param template:
        Enable templating of pillar files.  Defaults to ``False``.
    """
    # Not used
    del pillar

    if not root_dir:
        log.error("file_tree: no root_dir specified")
        return {}

    if not os.path.isabs(root_dir):
        pillarenv = __opts__["pillarenv"]
        if pillarenv is None:
            log.error("file_tree: root_dir is relative but pillarenv is not set")
            return {}
        log.debug("file_tree: pillarenv = %s", pillarenv)

        env_roots = __opts__["pillar_roots"].get(pillarenv, None)
        if env_roots is None:
            log.error(
                "file_tree: root_dir is relative but no pillar_roots are specified "
                " for pillarenv %s",
                pillarenv,
            )
            return {}

        env_dirs = []
        for env_root in env_roots:
            env_dir = os.path.normpath(os.path.join(env_root, root_dir))
            # don't redundantly load consecutively, but preserve any expected precedence
            if env_dir not in env_dirs or env_dir != env_dirs[-1]:
                env_dirs.append(env_dir)
        dirs = env_dirs
    else:
        dirs = [root_dir]

    result_pillar = {}
    for root in dirs:
        dir_pillar = _ext_pillar(
            minion_id,
            root,
            follow_dir_links,
            debug,
            keep_newline,
            render_default,
            renderer_blacklist,
            renderer_whitelist,
            template,
        )
        result_pillar = salt.utils.dictupdate.merge(
            result_pillar, dir_pillar, strategy="recurse"
        )
    return result_pillar


def _ext_pillar(
    minion_id,
    root_dir,
    follow_dir_links,
    debug,
    keep_newline,
    render_default,
    renderer_blacklist,
    renderer_whitelist,
    template,
):
    """
    Compile pillar data for a single root_dir for the specified minion ID
    """
    log.debug("file_tree: reading %s", root_dir)

    if not os.path.isdir(root_dir):
        log.error(
            "file_tree: root_dir %s does not exist or is not a directory", root_dir
        )
        return {}

    if not isinstance(keep_newline, (bool, list)):
        log.error(
            "file_tree: keep_newline must be either True/False or a list "
            "of file globs. Skipping this ext_pillar for root_dir %s",
            root_dir,
        )
        return {}

    ngroup_pillar = {}
    nodegroups_dir = os.path.join(root_dir, "nodegroups")
    if os.path.exists(nodegroups_dir) and len(__opts__.get("nodegroups", ())) > 0:
        master_ngroups = __opts__["nodegroups"]
        ext_pillar_dirs = os.listdir(nodegroups_dir)
        if len(ext_pillar_dirs) > 0:
            for nodegroup in ext_pillar_dirs:
                if os.path.isdir(nodegroups_dir) and nodegroup in master_ngroups:
                    ckminions = salt.utils.minions.CkMinions(__opts__)
                    _res = ckminions.check_minions(
                        master_ngroups[nodegroup], "compound"
                    )
                    match = _res["minions"]
                    if minion_id in match:
                        ngroup_dir = os.path.join(nodegroups_dir, str(nodegroup))
                        ngroup_pillar = salt.utils.dictupdate.merge(
                            ngroup_pillar,
                            _construct_pillar(
                                ngroup_dir,
                                follow_dir_links,
                                keep_newline,
                                render_default,
                                renderer_blacklist,
                                renderer_whitelist,
                                template,
                            ),
                            strategy="recurse",
                        )
        else:
            if debug is True:
                log.debug(
                    "file_tree: no nodegroups found in file tree directory %s,"
                    " skipping...",
                    ext_pillar_dirs,
                )
    else:
        if debug is True:
            log.debug("file_tree: no nodegroups found in master configuration")

    host_dir = os.path.join(root_dir, "hosts", minion_id)
    if not os.path.exists(host_dir):
        if debug is True:
            log.debug(
                "file_tree: no pillar data for minion %s found in file tree"
                " directory %s",
                minion_id,
                host_dir,
            )
        return ngroup_pillar

    if not os.path.isdir(host_dir):
        log.error("file_tree: %s exists, but is not a directory", host_dir)
        return ngroup_pillar

    host_pillar = _construct_pillar(
        host_dir,
        follow_dir_links,
        keep_newline,
        render_default,
        renderer_blacklist,
        renderer_whitelist,
        template,
    )
    return salt.utils.dictupdate.merge(ngroup_pillar, host_pillar, strategy="recurse")



"""
Splunk Search State Module

.. versionadded:: 2015.5.0

This state is used to ensure presence of splunk searches.

.. code-block:: yaml

    server-warning-message:
      splunk_search.present:
        - name: This is the splunk search name
        - search: index=main sourcetype=
"""


def __virtual__():
    """
    Only load if the splunk_search module is available in __salt__
    """
    if "splunk_search.get" in __salt__:
        return "splunk_search"
    return (False, "splunk module could not be loaded")


def present(name, profile="splunk", **kwargs):
    """
    Ensure a search is present

    .. code-block:: yaml

        API Error Search:
          splunk_search.present:
            search: index=main sourcetype=blah
            template: alert_5min

    The following parameters are required:

    name
        This is the name of the search in splunk
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    target = __salt__["splunk_search.get"](name, profile=profile)
    if target:
        if __opts__["test"]:
            ret["comment"] = f"Would update {name}"
            return ret
        # found a search... updating
        result = __salt__["splunk_search.update"](name, profile=profile, **kwargs)
        if not result:
            # no update
            ret["result"] = True
            ret["comment"] = "No changes"
        else:
            (newvalues, diffs) = result
            old_content = dict(target.content)
            old_changes = {}
            for x in newvalues:
                old_changes[x] = old_content.get(x, None)
            ret["result"] = True
            ret["changes"]["diff"] = diffs
            ret["changes"]["old"] = old_changes
            ret["changes"]["new"] = newvalues
    else:
        if __opts__["test"]:
            ret["comment"] = f"Would create {name}"
            return ret
        # creating a new search
        result = __salt__["splunk_search.create"](name, profile=profile, **kwargs)
        if result:
            ret["result"] = True
            ret["changes"]["old"] = False
            ret["changes"]["new"] = kwargs
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to create {name}"
    return ret


def absent(name, profile="splunk"):
    """
    Ensure a search is absent

    .. code-block:: yaml

        API Error Search:
          splunk_search.absent

    The following parameters are required:

    name
        This is the name of the search in splunk
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"{name} is absent.",
    }

    target = __salt__["splunk_search.get"](name, profile=profile)
    if target:
        if __opts__["test"]:
            ret = {}
            ret["name"] = name
            ret["comment"] = f"Would delete {name}"
            ret["result"] = None
            return ret

        result = __salt__["splunk_search.delete"](name, profile=profile)
        if result:
            ret["comment"] = f"{name} was deleted"
        else:
            ret["comment"] = f"Failed to delete {name}"
            ret["result"] = False
    return ret



"""
    :codeauthor: Marek Skrobacki <skrobul@skrobul.com>

    Test cases for salt.modules.s6
"""

import os

import pytest

import salt.modules.s6 as s6
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {s6: {"SERVICE_DIR": "/etc/service"}}


# 'start' function tests: 1


def test_start():
    """
    Test if it starts service via s6-svc.
    """
    mock_ret = MagicMock(return_value=False)
    with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
        assert s6.start("ssh")


# 'stop' function tests: 1


def test_stop():
    """
    Test if it stops service via s6.
    """
    mock_ret = MagicMock(return_value=False)
    with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
        assert s6.stop("ssh")


# 'term' function tests: 1


def test_term():
    """
    Test if it send a TERM to service via s6.
    """
    mock_ret = MagicMock(return_value=False)
    with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
        assert s6.term("ssh")


# 'reload_' function tests: 1


def test_reload():
    """
    Test if it send a HUP to service via s6.
    """
    mock_ret = MagicMock(return_value=False)
    with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
        assert s6.reload_("ssh")


# 'restart' function tests: 1


def test_restart():
    """
    Test if it restart service via s6. This will stop/start service.
    """
    mock_ret = MagicMock(return_value=False)
    with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
        assert s6.restart("ssh")


# 'full_restart' function tests: 1


def test_full_restart():
    """
    Test if it calls s6.restart() function.
    """
    mock_ret = MagicMock(return_value=False)
    with patch.dict(s6.__salt__, {"cmd.retcode": mock_ret}):
        assert s6.full_restart("ssh") is None


# 'status' function tests: 1


def test_status():
    """
    Test if it return the status for a service via s6,
    return pid if running.
    """
    mock_run = MagicMock(return_value="salt")
    with patch.dict(s6.__salt__, {"cmd.run_stdout": mock_run}):
        assert s6.status("ssh") == ""


# 'available' function tests: 1


def test_available():
    """
    Test if it returns ``True`` if the specified service is available,
    otherwise returns ``False``.
    """
    with patch.object(os, "listdir", MagicMock(return_value=["/etc/service"])):
        assert s6.available("/etc/service")


# 'missing' function tests: 1


def test_missing():
    """
    Test if it returns ``True`` if the specified service is not available,
    otherwise returns ``False``.
    """
    with patch.object(os, "listdir", MagicMock(return_value=["/etc/service"])):
        assert s6.missing("foo")


# 'get_all' function tests: 1


def test_get_all():
    """
    Test if it return a list of all available services.
    """
    with patch.object(os, "listdir", MagicMock(return_value=["/etc/service"])):
        assert s6.get_all() == ["/etc/service"]



"""
unit tests for the alias state
"""

import pytest

import salt.states.alias as alias
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {alias: {}}


def test_present_has_target():
    """
    test alias.present has target already
    """
    name = "saltdude"
    target = "dude@saltstack.com"
    ret = {
        "comment": f"Alias {name} already present",
        "changes": {},
        "name": name,
        "result": True,
    }

    has_target = MagicMock(return_value=True)
    with patch.dict(alias.__salt__, {"aliases.has_target": has_target}):
        assert alias.present(name, target) == ret


def test_present_has_not_target_test():
    """
    test alias.present has not target yet test mode
    """
    name = "saltdude"
    target = "dude@saltstack.com"
    ret = {
        "comment": f"Alias {name} -> {target} is set to be added",
        "changes": {},
        "name": name,
        "result": None,
    }

    has_target = MagicMock(return_value=False)
    with patch.dict(alias.__salt__, {"aliases.has_target": has_target}):
        with patch.dict(alias.__opts__, {"test": True}):
            assert alias.present(name, target) == ret


def test_present_set_target():
    """
    test alias.present set target
    """
    name = "saltdude"
    target = "dude@saltstack.com"
    ret = {
        "comment": f"Set email alias {name} -> {target}",
        "changes": {"alias": name},
        "name": name,
        "result": True,
    }

    has_target = MagicMock(return_value=False)
    set_target = MagicMock(return_value=True)
    with patch.dict(alias.__salt__, {"aliases.has_target": has_target}):
        with patch.dict(alias.__opts__, {"test": False}):
            with patch.dict(alias.__salt__, {"aliases.set_target": set_target}):
                assert alias.present(name, target) == ret


def test_present_set_target_failed():
    """
    test alias.present set target failure
    """
    name = "saltdude"
    target = "dude@saltstack.com"
    ret = {
        "comment": f"Failed to set alias {name} -> {target}",
        "changes": {},
        "name": name,
        "result": False,
    }

    has_target = MagicMock(return_value=False)
    set_target = MagicMock(return_value=False)
    with patch.dict(alias.__salt__, {"aliases.has_target": has_target}):
        with patch.dict(alias.__opts__, {"test": False}):
            with patch.dict(alias.__salt__, {"aliases.set_target": set_target}):
                assert alias.present(name, target) == ret


def test_absent_already_gone():
    """
    test alias.absent already gone
    """
    name = "saltdude"
    ret = {
        "comment": f"Alias {name} already absent",
        "changes": {},
        "name": name,
        "result": True,
    }

    get_target = MagicMock(return_value=False)
    with patch.dict(alias.__salt__, {"aliases.get_target": get_target}):
        assert alias.absent(name) == ret


def test_absent_not_gone_test():
    """
    test alias.absent already gone test mode
    """
    name = "saltdude"
    ret = {
        "comment": f"Alias {name} is set to be removed",
        "changes": {},
        "name": name,
        "result": None,
    }

    get_target = MagicMock(return_value=True)
    with patch.dict(alias.__salt__, {"aliases.get_target": get_target}):
        with patch.dict(alias.__opts__, {"test": True}):
            assert alias.absent(name) == ret


def test_absent_rm_alias():
    """
    test alias.absent remove alias
    """
    name = "saltdude"
    ret = {
        "comment": f"Removed alias {name}",
        "changes": {"alias": name},
        "name": name,
        "result": True,
    }

    get_target = MagicMock(return_value=True)
    rm_alias = MagicMock(return_value=True)
    with patch.dict(alias.__salt__, {"aliases.get_target": get_target}):
        with patch.dict(alias.__opts__, {"test": False}):
            with patch.dict(alias.__salt__, {"aliases.rm_alias": rm_alias}):
                assert alias.absent(name) == ret


def test_absent_rm_alias_failed():
    """
    test alias.absent remove alias failure
    """
    name = "saltdude"
    ret = {
        "comment": f"Failed to remove alias {name}",
        "changes": {},
        "name": name,
        "result": False,
    }

    get_target = MagicMock(return_value=True)
    rm_alias = MagicMock(return_value=False)
    with patch.dict(alias.__salt__, {"aliases.get_target": get_target}):
        with patch.dict(alias.__opts__, {"test": False}):
            with patch.dict(alias.__salt__, {"aliases.rm_alias": rm_alias}):
                assert alias.absent(name) == ret



"""
Tests for the salt fileclient
"""

import errno
import logging
import os

import pytest

import salt.utils.files
from salt import fileclient
from tests.support.mock import AsyncMock, MagicMock, Mock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def mocked_opts(tmp_path):
    fs_root = os.path.join(tmp_path, "fileclient_fs_root")
    cache_root = os.path.join(tmp_path, "fileclient_cache_root")
    return {
        "file_roots": {x: [os.path.join(fs_root, x)] for x in ("base", "dev")},
        "fileserver_backend": ["roots"],
        "cachedir": cache_root,
        "file_client": "local",
    }


@pytest.fixture
def configure_loader_modules(tmp_path, mocked_opts):
    return {fileclient: {"__opts__": mocked_opts}}


@pytest.fixture
def file_client(mocked_opts):
    client = fileclient.Client(mocked_opts)
    try:
        yield client
    finally:
        del client


@pytest.fixture
def client_opts():
    return {
        "extension_modules": "",
        "cachedir": "/__test__",
    }


def _fake_makedir(num=errno.EEXIST):
    def _side_effect(*args, **kwargs):
        raise OSError(num, f"Errno {num}")

    return Mock(side_effect=_side_effect)


class MockReqChannel:
    def factory(self, opts):
        return self

    def close(self):
        return True

    def send(self, load):
        return self


def test_fileclient_context_manager_closes(minion_opts, master_opts):
    """
    ensure fileclient channel closes
    when used with a context manager
    """
    minion_opts.update(
        {
            "id": "root",
            "transport": "zeromq",
            "auth_tries": 1,
            "auth_timeout": 5,
            "master_ip": "127.0.0.1",
            "master_port": master_opts["ret_port"],
            "master_uri": f"tcp://127.0.0.1:{master_opts['ret_port']}",
            "request_channel_timeout": 1,
            "request_channel_tries": 1,
        }
    )
    mock_reqchannel = MockReqChannel()
    patch_reqchannel = patch.object(
        salt.channel.client, "ReqChannel", return_value=mock_reqchannel
    )
    with patch_reqchannel:
        with fileclient.get_file_client(minion_opts) as client:
            client.master_opts()
            assert not client._closing

        assert client._closing
        assert client.channel.close.called


@pytest.mark.slow_test
def test_fileclient_timeout(minion_opts, master_opts):
    """
    ensure fileclient channel closes
    when used with a context manager
    """
    minion_opts.update(
        {
            "id": "root",
            "transport": "zeromq",
            "auth_tries": 1,
            "auth_timeout": 5,
            "master_ip": "127.0.0.1",
            "master_port": master_opts["ret_port"],
            "master_uri": f"tcp://127.0.0.1:{master_opts['ret_port']}",
            "request_channel_timeout": 1,
            "request_channel_tries": 1,
        }
    )

    def mock_dumps(*args):
        return b"meh"

    with fileclient.get_file_client(minion_opts) as client:
        # Authenticate must return true
        with patch.object(client.auth, "authenticate", AsyncMock(return_value=True)):
            # Crypticle must return bytes to pass to transport.RequestClient.send
            client.auth._crypticle = Mock()
            client.auth._crypticle.dumps = mock_dumps
            msg = r"^File client timed out after \d{1,4} seconds$"
            with pytest.raises(salt.exceptions.SaltClientError, match=msg):
                client.file_list()


def test_cache_skips_makedirs_on_race_condition(client_opts):
    """
    If cache contains already a directory, do not raise an exception.
    """
    with patch("os.path.isfile", return_value=False):
        for exists in range(2):
            with patch("os.makedirs", _fake_makedir()):
                with fileclient.Client(client_opts)._cache_loc("testfile") as c_ref_itr:
                    assert c_ref_itr == os.sep + os.sep.join(
                        ["__test__", "files", "base", "testfile"]
                    )


def test_cache_raises_exception_on_non_eexist_ioerror(client_opts):
    """
    If makedirs raises other than EEXIST errno, an exception should be raised.
    """
    with patch("os.path.isfile", return_value=False):
        with patch("os.makedirs", _fake_makedir(num=errno.EROFS)):
            with pytest.raises(OSError):
                with fileclient.Client(client_opts)._cache_loc("testfile") as c_ref_itr:
                    assert c_ref_itr == "/__test__/files/base/testfile"


def test_cache_extrn_path_valid(client_opts):
    """
    Tests for extrn_filepath for a given url
    """
    file_name = "http://localhost:8000/test/location/src/dev/usr/file"

    ret = fileclient.Client(client_opts)._extrn_path(file_name, "base")
    assert ":" not in ret
    assert ret == os.path.join("__test__", "extrn_files", "base", ret)


def test_cache_extrn_path_invalid(client_opts):
    """
    Tests for extrn_filepath for a given url
    """
    file_name = "http://localhost:8000/../../../../../usr/bin/bad"

    ret = fileclient.Client(client_opts)._extrn_path(file_name, "base")
    assert ret == "Invalid path"


def test_extrn_path_with_long_filename(client_opts):
    safe_file_name = os.path.split(
        fileclient.Client(client_opts)._extrn_path(
            "https://test.com/" + ("A" * 254), "base"
        )
    )[-1]
    assert safe_file_name == "A" * 254

    oversized_file_name = os.path.split(
        fileclient.Client(client_opts)._extrn_path(
            "https://test.com/" + ("A" * 255), "base"
        )
    )[-1]
    assert len(oversized_file_name) < 256
    assert oversized_file_name != "A" * 255

    oversized_file_with_query_params = os.path.split(
        fileclient.Client(client_opts)._extrn_path(
            "https://test.com/file?" + ("A" * 255), "base"
        )
    )[-1]
    assert len(oversized_file_with_query_params) < 256


def test_file_list_emptydirs(file_client):
    """
    Ensure that the fileclient class won't allow a direct call to file_list_emptydirs()
    """
    with pytest.raises(NotImplementedError):
        file_client.file_list_emptydirs()


def test_get_file(file_client):
    """
    Ensure that the fileclient class won't allow a direct call to get_file()
    """
    with pytest.raises(NotImplementedError):
        file_client.get_file(None)


def test_get_file_client(file_client):
    minion_opts = {}
    minion_opts["file_client"] = "remote"
    with patch("salt.fileclient.RemoteClient", MagicMock(return_value="remote_client")):
        ret = fileclient.get_file_client(minion_opts)
        assert "remote_client" == ret


def test_get_url_with_hash(client_opts):
    """
    Test get_url function with a URL containing a hash character.
    """
    with patch("os.path.isfile", return_value=True):
        with patch("os.makedirs", return_value=None):
            with patch("urllib.request.urlretrieve", return_value=None):
                client = fileclient.Client(client_opts)
                url = "file:///path/to/file#with#hash"
                dest = "/mocked/destination"

                result = client.get_url(url, dest)

                assert result == "/path/to/file#with#hash"



"""
    :codeauthor: Denys Havrysh <denys.gavrysh@gmail.com>
"""

import logging
import os
import pprint

import pytest

import salt._logging
import salt.config
import salt.syspaths
import salt.utils.jid
import salt.utils.parsers
import salt.utils.platform
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


class LogImplMock:
    """
    Logger setup
    """

    def __init__(self):
        """
        init
        """
        self.log_level_console = None
        self.log_file = None
        self.log_level_logfile = None
        self.config = self.original_config = None
        logging_options = salt._logging.get_logging_options_dict()
        if logging_options:
            self.config = logging_options.copy()
            self.original_config = self.config.copy()
        self.temp_log_level = None
        self._console_handler_configured = False
        self._extended_logging_configured = False
        self._logfile_handler_configured = False
        self._real_set_logging_options_dict = salt._logging.set_logging_options_dict
        self._real_get_logging_options_dict = salt._logging.get_logging_options_dict
        self._real_setup_logfile_handler = salt._logging.setup_logfile_handler

    def _destroy(self):
        salt._logging.set_logging_options_dict.__options_dict__ = self.original_config
        salt._logging.shutdown_logfile_handler()

    def setup_temp_handler(self, log_level=None):
        """
        Set temp handler loglevel
        """
        log.debug("Setting temp handler log level to: %s", log_level)
        self.temp_log_level = log_level

    def is_console_handler_configured(self):
        log.debug("Calling is_console_handler_configured")
        return self._console_handler_configured

    def setup_console_handler(
        self, log_level="error", **kwargs
    ):  # pylint: disable=unused-argument
        """
        Set console loglevel
        """
        log.debug("Setting console handler log level to: %s", log_level)
        self.log_level_console = log_level
        self._console_handler_configured = True

    def shutdown_console_handler(self):
        log.debug("Calling shutdown_console_handler")
        self._console_handler_configured = False

    def is_extended_logging_configured(self):
        log.debug("Calling is_extended_logging_configured")
        return self._extended_logging_configured

    def setup_extended_logging(self, opts):
        """
        Set opts
        """
        log.debug("Calling setup_extended_logging")
        self._extended_logging_configured = True

    def shutdown_extended_logging(self):
        log.debug("Calling shutdown_extended_logging")
        self._extended_logging_configured = False

    def is_logfile_handler_configured(self):
        log.debug("Calling is_logfile_handler_configured")
        return self._logfile_handler_configured

    def setup_logfile_handler(
        self, log_path, log_level=None, **kwargs
    ):  # pylint: disable=unused-argument
        """
        Set logfile and loglevel
        """
        log.debug("Setting log file handler path to: %s", log_path)
        log.debug("Setting log file handler log level to: %s", log_level)
        self.log_file = log_path
        self.log_level_logfile = log_level
        self._real_setup_logfile_handler(log_path, log_level=log_level, **kwargs)
        self._logfile_handler_configured = True

    def shutdown_logfile_handler(self):
        log.debug("Calling shutdown_logfile_handler")
        self._logfile_handler_configured = False

    def get_logging_options_dict(self):
        log.debug("Calling get_logging_options_dict")
        return self.config

    def set_logging_options_dict(self, opts):
        log.debug("Calling set_logging_options_dict")
        self._real_set_logging_options_dict(opts)
        self.config = self._real_get_logging_options_dict()
        log.debug("Logging options dict:\n%s", pprint.pformat(self.config))

    def setup_log_granular_levels(self, opts):
        log.debug("Calling setup_log_granular_levels")

    def setup_logging(self):
        log.debug("Mocked setup_logging called")
        # Wether daemonizing or not, either on the main process or on a separate process
        # The log file is going to be configured.
        # The console is the only handler not configured if daemonizing

        # These routines are what happens on salt._logging.setup_logging
        opts = self.get_logging_options_dict()

        if (
            opts.get("configure_console_logger", True)
            and not self.is_console_handler_configured()
        ):
            self.setup_console_handler(
                log_level=opts["log_level"],
                log_format=opts["log_fmt_console"],
                date_format=opts["log_datefmt"],
            )
        if (
            opts.get("configure_file_logger", True)
            and not self.is_logfile_handler_configured()
        ):
            log_file_level = opts["log_level_logfile"] or opts["log_level"]
            if log_file_level != "quiet":
                self.setup_logfile_handler(
                    log_path=opts[opts["log_file_key"]],
                    log_level=log_file_level,
                    log_format=opts["log_fmt_logfile"],
                    date_format=opts["log_datefmt_logfile"],
                    max_bytes=opts["log_rotate_max_bytes"],
                    backup_count=opts["log_rotate_backup_count"],
                    user=opts["user"],
                )
        if not self.is_extended_logging_configured():
            self.setup_extended_logging(opts)
        self.setup_log_granular_levels(opts["log_granular_levels"])

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self._destroy()


# <----------- START TESTS ----------->


@pytest.fixture
def root_dir(tmp_path):
    yield tmp_path / "parsers_tests_root_dir"


@pytest.fixture(
    params=[
        "master",
        "minion",
        "proxyminion",
        "syndic",
        "saltcmd",
        "saltcp",
        "saltkey",
        "saltcall",
        "saltrun",
        "saltssh",
        "saltcloud",
        "spm",
        "saltapi",
    ]
)
def log_cli_parser(request):
    return request.param


@pytest.fixture
def default_config(log_cli_parser):
    if log_cli_parser == "master":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "minion":
        return salt.config.DEFAULT_MINION_OPTS.copy()
    elif log_cli_parser == "proxyminion":
        return {
            **salt.config.DEFAULT_MINION_OPTS.copy(),
            **salt.config.DEFAULT_PROXY_MINION_OPTS.copy(),
        }
    elif log_cli_parser == "syndic":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "saltcmd":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "saltcp":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "saltkey":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "saltcall":
        return salt.config.DEFAULT_MINION_OPTS.copy()
    elif log_cli_parser == "saltrun":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "saltssh":
        return salt.config.DEFAULT_MASTER_OPTS.copy()
    elif log_cli_parser == "saltcloud":
        return {
            **salt.config.DEFAULT_MASTER_OPTS.copy(),
            **salt.config.DEFAULT_CLOUD_OPTS.copy(),
        }
    elif log_cli_parser == "spm":
        return {
            **salt.config.DEFAULT_MASTER_OPTS.copy(),
            **salt.config.DEFAULT_SPM_OPTS.copy(),
        }
    elif log_cli_parser == "saltapi":
        return {
            **salt.config.DEFAULT_MASTER_OPTS.copy(),
            **salt.config.DEFAULT_API_OPTS.copy(),
        }


@pytest.fixture
def parser(log_cli_parser):
    param_map = {
        "master": salt.utils.parsers.MasterOptionParser,
        "minion": salt.utils.parsers.MinionOptionParser,
        "proxyminion": salt.utils.parsers.ProxyMinionOptionParser,
        "syndic": salt.utils.parsers.SyndicOptionParser,
        "saltcmd": salt.utils.parsers.SaltCMDOptionParser,
        "saltcp": salt.utils.parsers.SaltCPOptionParser,
        "saltkey": salt.utils.parsers.SaltKeyOptionParser,
        "saltcall": salt.utils.parsers.SaltCallOptionParser,
        "saltrun": salt.utils.parsers.SaltRunOptionParser,
        "saltssh": salt.utils.parsers.SaltSSHOptionParser,
        "saltcloud": salt.utils.parsers.SaltCloudParser,
        "spm": salt.utils.parsers.SPMParser,
        "saltapi": salt.utils.parsers.SaltAPIParser,
    }
    return param_map[log_cli_parser]


@pytest.fixture
def config_func(log_cli_parser):
    param_map = {
        "master": "salt.config.master_config",
        "minion": "salt.config.minion_config",
        "proxyminion": "salt.config.proxy_config",
        "syndic": "salt.config.syndic_config",
        "saltcmd": "salt.config.client_config",
        "saltcp": "salt.config.master_config",
        "saltkey": "salt.config.client_config",
        "saltcall": "salt.config.minion_config",
        "saltrun": "salt.config.master_config",
        "saltssh": "salt.config.master_config",
        "saltcloud": "salt.config.cloud_config",
        "spm": "salt.config.spm_config",
        "saltapi": "salt.config.api_config",
    }
    return param_map[log_cli_parser]


@pytest.fixture
def log_file(tmp_path, logfile_config_setting_name):
    return str(tmp_path / logfile_config_setting_name)


@pytest.fixture
def args(log_cli_parser):
    if log_cli_parser in ("saltcmd", "saltssh"):
        return ["foo", "bar.baz"]
    elif log_cli_parser == "saltcp":
        return ["foo", "bar", "baz"]
    elif log_cli_parser in ("saltcall", "saltrun"):
        return ["foo.bar"]
    elif log_cli_parser == "saltcloud":
        return ["-p", "foo", "bar"]
    elif log_cli_parser == "spm":
        return ["foo", "bar"]
    return []


@pytest.fixture
def loglevel_config_setting_name():
    return "log_level"


@pytest.fixture
def logfile_config_setting_name(log_cli_parser):
    if log_cli_parser == "syndic":
        return "syndic_log_file"
    elif log_cli_parser == "saltkey":
        return "key_logfile"
    elif log_cli_parser == "saltssh":
        return "ssh_log_file"
    elif log_cli_parser == "spm":
        return "spm_logfile"
    elif log_cli_parser == "saltapi":
        return "api_logfile"
    return "log_file"


@pytest.fixture
def logfile_loglevel_config_setting_name():
    return "log_level_logfile"


@pytest.fixture
def testing_config(default_config, root_dir, logfile_config_setting_name, log_file):
    _testing_config = default_config.copy()
    _testing_config["root_dir"] = root_dir
    for name in ("pki_dir", "cachedir"):
        _testing_config[name] = name
    _testing_config[logfile_config_setting_name] = log_file
    return _testing_config


@pytest.fixture(autouse=True)
def log_impl():
    """
    Mock logger functions
    """
    with LogImplMock() as _log_impl:
        mocked_functions = {}
        for name in dir(_log_impl):
            if name.startswith("_"):
                continue
            func = getattr(_log_impl, name)
            if not callable(func):
                continue
            mocked_functions[name] = func

        patcher = patch.multiple(salt._logging, **mocked_functions)
        with patcher:
            yield _log_impl


def test_get_log_level_cli(
    testing_config, loglevel_config_setting_name, args, parser, config_func, log_impl
):
    """
    Tests that log level match command-line specified value
    """
    # Set defaults
    default_log_level = testing_config[loglevel_config_setting_name]

    # Set log level in CLI
    log_level = "critical"
    args = ["--log-level", log_level] + args

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    console_log_level = getattr(instance.options, loglevel_config_setting_name)

    # Check console log level setting
    assert console_log_level == log_level
    # Check console logger log level
    assert log_impl.log_level_console == log_level
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.temp_log_level == log_level
    # Check log file logger log level
    assert log_impl.log_level_logfile == default_log_level


def test_get_log_level_config(
    testing_config, loglevel_config_setting_name, args, parser, config_func, log_impl
):
    """
    Tests that log level match the configured value
    """
    # Set log level in config
    log_level = "info"
    testing_config.update({loglevel_config_setting_name: log_level})

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    console_log_level = getattr(instance.options, loglevel_config_setting_name)

    # Check console log level setting
    assert console_log_level == log_level
    # Check console logger log level
    assert log_impl.log_level_console == log_level
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.temp_log_level == "error"
    # Check log file logger log level
    assert log_impl.log_level_logfile == log_level


def test_get_log_level_default(
    testing_config, loglevel_config_setting_name, args, parser, config_func, log_impl
):
    """
    Tests that log level match the default value
    """
    # Set defaults
    log_level = default_log_level = testing_config[loglevel_config_setting_name]

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    console_log_level = getattr(instance.options, loglevel_config_setting_name)

    # Check log level setting
    assert console_log_level == log_level
    # Check console logger log level
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.temp_log_level == "error"
    # Check log file logger
    assert log_impl.log_level_logfile == default_log_level
    # Check help message
    assert f"Default: '{default_log_level}'." in instance.get_option("--log-level").help


# log file configuration tests


def test_get_log_file_cli(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    log_file,
    logfile_config_setting_name,
):
    """
    Tests that log file match command-line specified value
    """
    # Set defaults
    log_level = testing_config[loglevel_config_setting_name]

    # Set log file in CLI
    log_file = f"{log_file}_cli.log"
    args = ["--log-file", log_file] + args

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_file_option = getattr(instance.options, logfile_config_setting_name)

    # Check console logger
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.config[logfile_config_setting_name] == log_file
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file setting
    assert log_file_option == log_file
    # Check log file logger
    assert log_impl.log_file == log_file


def test_get_log_file_config(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    logfile_config_setting_name,
    log_file,
):
    """
    Tests that log file match the configured value
    """
    # Set defaults
    log_level = testing_config[loglevel_config_setting_name]

    # Set log file in config
    log_file = f"{log_file}_config.log"
    testing_config.update({logfile_config_setting_name: log_file})

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_file_option = getattr(instance.options, logfile_config_setting_name)

    # Check console logger
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.config[logfile_config_setting_name] == log_file
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file setting
    assert log_file_option == log_file
    # Check log file logger
    assert log_impl.log_file == log_file


def test_get_log_file_default(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    logfile_config_setting_name,
    default_config,
):
    """
    Tests that log file match the default value
    """
    # Set defaults
    log_level = testing_config[loglevel_config_setting_name]
    log_file = testing_config[logfile_config_setting_name]
    default_log_file = default_config[logfile_config_setting_name]

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_file_option = getattr(instance.options, logfile_config_setting_name)

    # Check console logger
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.config[logfile_config_setting_name] == log_file
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file setting
    assert log_file_option == log_file
    # Check log file logger
    assert log_impl.log_file == log_file
    # Check help message
    assert f"Default: '{default_log_file}'." in instance.get_option("--log-file").help


# log file log level configuration tests


def test_get_log_file_level_cli(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    logfile_loglevel_config_setting_name,
):
    """
    Tests that file log level match command-line specified value
    """
    # Set defaults
    default_log_level = testing_config[loglevel_config_setting_name]

    # Set log file level in CLI
    log_level_logfile = "error"
    args = ["--log-file-level", log_level_logfile] + args

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_level_logfile_option = getattr(
        instance.options, logfile_loglevel_config_setting_name
    )

    # Check console logger
    assert log_impl.log_level_console == default_log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == default_log_level
    assert log_impl.config[logfile_loglevel_config_setting_name] == log_level_logfile
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file level setting
    assert log_level_logfile_option == log_level_logfile
    # Check log file logger
    assert log_impl.log_level_logfile == log_level_logfile


def test_get_log_file_level_config(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    logfile_loglevel_config_setting_name,
):
    """
    Tests that log file level match the configured value
    """
    # Set defaults
    log_level = testing_config[loglevel_config_setting_name]

    # Set log file level in config
    log_level_logfile = "info"
    testing_config.update({logfile_loglevel_config_setting_name: log_level_logfile})

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_level_logfile_option = getattr(
        instance.options, logfile_loglevel_config_setting_name
    )

    # Check console logger
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.config[logfile_loglevel_config_setting_name] == log_level_logfile
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file level setting
    assert log_level_logfile_option == log_level_logfile
    # Check log file logger
    assert log_impl.log_level_logfile == log_level_logfile


def test_get_log_file_level_default(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    logfile_loglevel_config_setting_name,
):
    """
    Tests that log file level match the default value
    """
    # Set defaults
    default_log_level = testing_config[loglevel_config_setting_name]

    log_level = default_log_level
    log_level_logfile = default_log_level

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_level_logfile_option = getattr(
        instance.options, logfile_loglevel_config_setting_name
    )

    # Check console logger
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.config[logfile_loglevel_config_setting_name] == log_level_logfile
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file level setting
    assert log_level_logfile_option == log_level_logfile
    # Check log file logger
    assert log_impl.log_level_logfile == log_level_logfile
    # Check help message
    assert (
        f"Default: '{default_log_level}'."
        in instance.get_option("--log-file-level").help
    )


def test_get_console_log_level_with_file_log_level(
    testing_config,
    loglevel_config_setting_name,
    args,
    parser,
    config_func,
    log_impl,
    logfile_loglevel_config_setting_name,
):  # pylint: disable=invalid-name
    """
    Tests that both console log level and log file level setting are working together
    """
    log_level = "critical"
    log_level_logfile = "debug"

    args = ["--log-file-level", log_level_logfile] + args

    testing_config.update({loglevel_config_setting_name: log_level})

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    log_level_logfile_option = getattr(
        instance.options, logfile_loglevel_config_setting_name
    )

    # Check console logger
    assert log_impl.log_level_console == log_level
    # Check extended logger
    assert log_impl.config[loglevel_config_setting_name] == log_level
    assert log_impl.config[logfile_loglevel_config_setting_name] == log_level_logfile
    # Check temp logger
    assert log_impl.temp_log_level == "error"
    # Check log file level setting
    assert log_level_logfile_option == log_level_logfile
    # Check log file logger
    assert log_impl.log_level_logfile == log_level_logfile


def test_log_created(
    testing_config, args, parser, config_func, logfile_config_setting_name, log_file
):
    """
    Tests that log file is created
    """
    testing_config.update({"log_file": str(log_file)})
    log_file_name = str(log_file)
    if log_file_name.rsplit(os.sep, maxsplit=1)[-1] != "log_file":
        testing_config.update({log_file_name: str(log_file)})

    instance = parser()
    with patch(config_func, MagicMock(return_value=testing_config)):
        instance.parse_args(args)

    assert os.path.exists(str(log_file_name))


def test_callbacks_uniqueness(parser):
    """
    Test that the callbacks are only added once, no matter
    how many instances of the parser we create
    """
    mixin_container_names = (
        "_mixin_setup_funcs",
        "_mixin_process_funcs",
        "_mixin_after_parsed_funcs",
        "_mixin_before_exit_funcs",
    )
    instance = parser()
    nums_1 = {}
    for cb_container in mixin_container_names:
        obj = getattr(instance, cb_container)
        nums_1[cb_container] = len(obj)

    # The next time we instantiate the parser, the counts should be equal
    instance = parser()
    nums_2 = {}
    for cb_container in mixin_container_names:
        obj = getattr(instance, cb_container)
        nums_2[cb_container] = len(obj)
    assert nums_1 == nums_2


def test_verify_log_warning_logged(args, config_func, testing_config, parser, caplog):
    args = ["--log-level", "debug"] + args
    with caplog.at_level(logging.DEBUG):
        instance = parser()
        with patch(config_func, MagicMock(return_value=testing_config)):
            instance.parse_args(args)
        assert (
            "Insecure logging configuration detected! Sensitive data may be logged."
            in caplog.messages
        )



import ctypes
import hashlib
import logging
import multiprocessing
import queue
import socket
import time

import pytest
import tornado.ioloop
import tornado.iostream
import zmq
from pytestshellutils.utils.processes import terminate_process

import salt.channel.server
import salt.crypt
import salt.exceptions
import salt.master
import salt.payload
import salt.utils.msgpack
import salt.utils.process
import salt.utils.stringutils

log = logging.getLogger(__name__)


class RecvError(Exception):
    """
    Raised by the Collector's _recv method when there is a problem
    getting publishes from to the publisher.
    """


class Collector(salt.utils.process.SignalHandlingProcess):
    def __init__(
        self,
        minion_config,
        interface,
        port,
        aes_key,
        timeout=300,
    ):
        super().__init__()
        self.minion_config = minion_config
        self.hexid = hashlib.sha1(
            salt.utils.stringutils.to_bytes(self.minion_config["id"])
        ).hexdigest()
        self.interface = interface
        self.port = port
        self.aes_key = aes_key
        self.timeout = timeout
        self.aes_key = aes_key
        self.hard_timeout = time.time() + timeout + 120
        self.manager = multiprocessing.Manager()
        self.results = self.manager.list()
        self.zmq_filtering = minion_config["zmq_filtering"]
        self.stopped = multiprocessing.Event()
        self.started = multiprocessing.Event()
        self.running = multiprocessing.Event()
        self.stop_running = multiprocessing.Event()
        self.unpacker = salt.utils.msgpack.Unpacker(raw=False)

    @property
    def transport(self):
        return self.minion_config["transport"]

    def _rotate_secrets(self, now=None):
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
            "reload": salt.crypt.Crypticle.generate_key_string,
            "rotate_master_key": self._rotate_secrets,
        }

    def _teardown_listener(self):
        if self.transport == "zeromq":
            self.sock.close()
            self.ctx.term()
        else:
            self.sock.close()

    def _setup_listener(self):
        if self.transport == "zeromq":
            self.ctx = zmq.Context()
            self.sock = self.ctx.socket(zmq.SUB)
            self.sock.setsockopt(zmq.LINGER, -1)
            if self.zmq_filtering:
                self.sock.setsockopt(zmq.SUBSCRIBE, b"broadcast")
                if self.minion_config.get("__role") == "syndic":
                    self.sock.setsockopt(zmq.SUBSCRIBE, b"syndic")
                else:
                    self.sock.setsockopt(
                        zmq.SUBSCRIBE, salt.utils.stringutils.to_bytes(self.hexid)
                    )
            else:
                self.sock.setsockopt(zmq.SUBSCRIBE, b"")
            pub_uri = f"tcp://{self.interface}:{self.port}"
            self.sock.connect(pub_uri)
        else:
            end = time.time() + 120
            while True:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    sock.connect((self.interface, self.port))
                except ConnectionRefusedError:
                    if time.time() >= end:
                        raise
                    time.sleep(1)
                else:
                    break
            self.sock = tornado.iostream.IOStream(sock)

    async def _recv(self):
        if self.transport == "zeromq":
            # test_zeromq_filtering requires catching the
            # SaltDeserializationError in order to pass.
            try:
                messages = self.sock.recv_multipart(flags=zmq.NOBLOCK, copy=True)
                messages_len = len(messages)
                if messages_len == 1:
                    serial_payload = salt.payload.loads(messages[0])
                elif messages_len == 2:
                    message_target = salt.utils.stringutils.to_str(messages[0])
                    is_syndic = self.minion_config.get("__role") == "syndic"
                    if (
                        not is_syndic
                        and message_target not in ("broadcast", self.hexid)
                    ) or (is_syndic and message_target not in ("broadcast", "syndic")):
                        log.debug(
                            "Publish received for not this minion: %s", message_target
                        )
                        return
                    serial_payload = salt.payload.loads(messages[1])
                else:
                    raise Exception("Invalid number of messages")
                return serial_payload
            except (zmq.ZMQError, salt.exceptions.SaltDeserializationError):
                raise RecvError("ZMQ Error")
        else:
            for msg in self.unpacker:
                serial_payload = salt.payload.loads(msg["body"])
                return serial_payload
            byts = await self.sock.read_bytes(8096, partial=True)
            self.unpacker.feed(byts)
            for msg in self.unpacker:
                serial_payload = salt.payload.loads(msg["body"])
                return serial_payload
            raise RecvError("TCP Error")

    async def _run(self, loop):
        try:
            self._setup_listener()
        except Exception:  # pylint: disable=broad-except
            self.started.set()
            log.exception("Failed to start listening")
            return
        self.started.set()
        last_msg = time.time()
        self.start = last_msg
        crypticle = salt.crypt.Crypticle(self.minion_config, self.aes_key)
        self.gotone = False
        try:
            while True:
                curr_time = time.time()
                if time.time() > self.hard_timeout:
                    log.error("Hard timeout reached in test collector!")
                    break
                if curr_time - last_msg >= self.timeout:
                    log.error("Receive timeout reached in test collector!")
                    break
                try:
                    payload = await self._recv()
                except RecvError:
                    time.sleep(0.03)
                else:
                    try:
                        log.trace("Colleted payload %r", payload)
                        payload = crypticle.loads(payload["load"])
                        if not payload:
                            continue
                        if "start" in payload:
                            log.info("Collector started")
                            self.running.set()
                            continue
                        if "stop" in payload:
                            log.info("Collector stopped")
                            self.stop_running.set()
                            break
                        last_msg = time.time()
                        self.results.append(payload["jid"])
                    except salt.exceptions.SaltDeserializationError:
                        log.error("Deserializer Error")
                        if not self.zmq_filtering:
                            log.exception("Failed to deserialize...")
                            break
                    if self.gotone is False:
                        log.debug("Collector started recieving")
                    self.gotone = True
            log.debug("Collector finished recieving")
            self.end = time.time()
            print(f"Total time {self.end - self.start}")
        finally:
            self._teardown_listener()
            loop.stop()

    def run(self):
        """
        Gather results until then number of seconds specified by timeout passes
        without receiving a message
        """
        loop = tornado.ioloop.IOLoop()
        loop.add_callback(self._run, loop)
        loop.start()

    def __enter__(self):
        self.manager.__enter__()
        self.start()
        # Wait until we can start receiving events
        self.started.wait()
        self.started.clear()
        return self

    def __exit__(self, *args):
        # Wait until we either processed all expected messages or we reach the hard timeout
        join_secs = self.hard_timeout - time.time()
        log.info("Waiting at most %s seconds before exiting the collector", join_secs)
        self.join(join_secs)
        self.terminate()
        # Cast our manager.list into a plain list
        self.results = list(self.results)
        # Terminate our multiprocessing manager
        self.manager.__exit__(*args)
        log.debug("The collector has exited")
        self.stopped.set()


class PubServerChannelProcess(salt.utils.process.SignalHandlingProcess):
    def __init__(self, master_config, minion_config, **collector_kwargs):
        super().__init__()
        self.name = "PubServerChannelProcess"
        self._closing = False
        self.master_config = master_config
        self.minion_config = minion_config
        self.collector_kwargs = collector_kwargs
        self.aes_key = salt.crypt.Crypticle.generate_key_string()
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(self.aes_key),
            ),
            "serial": multiprocessing.Value(
                ctypes.c_longlong, lock=False  # We'll use the lock from 'secret'
            ),
        }
        self.process_manager = salt.utils.process.ProcessManager(
            name="ZMQ-PubServer-ProcessManager"
        )
        self.pub_server_channel = salt.channel.server.PubServerChannel.factory(
            self.master_config
        )
        self.pub_server_channel.pre_fork(self.process_manager)
        self.pub_uri = "tcp://{interface}:{publish_port}".format(**self.master_config)
        self.queue = multiprocessing.Queue()
        self.stopped = multiprocessing.Event()
        self.collector = Collector(
            self.minion_config,
            self.master_config["interface"],
            self.master_config["publish_port"],
            self.aes_key,
            **self.collector_kwargs,
        )

    def run(self):
        ioloop = tornado.ioloop.IOLoop()
        try:
            while True:
                try:
                    payload = self.queue.get(False)
                except queue.Empty:
                    if self._closing is True:
                        break
                    time.sleep(0.03)
                    continue
                if payload is None:
                    log.debug("We received the stop sentinel")
                    break
                ioloop.run_sync(lambda: self.pub_server_channel.publish(payload))
        except KeyboardInterrupt:
            pass
        finally:
            self.stopped.set()

    def _handle_signals(self, signum, sigframe):
        self.close()
        super()._handle_signals(signum, sigframe)

    def close(self):
        if self._closing:
            return
        self._closing = True
        if self.process_manager is None:
            return
        self.process_manager.terminate()
        if hasattr(self.pub_server_channel, "close"):
            self.pub_server_channel.close()
        # Really terminate any process still left behind
        for pid in self.process_manager._process_map:
            terminate_process(pid=pid, kill_children=True, slow_stop=False)
        self.process_manager = None

    def publish(self, payload):
        self.queue.put(payload)

    def __enter__(self):
        self.start()
        self.collector.__enter__()
        attempts = 300
        while attempts > 0:
            self.publish({"tgt_type": "glob", "tgt": "*", "jid": -1, "start": True})
            if self.collector.running.wait(1) is True:
                break
            time.sleep(0.5)
            attempts -= 1
        else:
            pytest.fail("Failed to confirm the collector has started")
        return self

    def __exit__(self, *args):
        # Publish a payload to tell the collection it's done processing
        attempts = 300
        while attempts > 0:
            self.publish({"tgt_type": "glob", "tgt": "*", "jid": -1, "stop": True})
            if self.collector.stop_running.wait(1) is True:
                break
            time.sleep(0.5)
            attempts -= 1
        else:
            pytest.fail("Failed to confirm the collector has stopped")
        # Now trigger the collector to also exit
        self.collector.__exit__(*args)
        # We can safely wait here without a timeout because the Collector instance has a
        # hard timeout set, so eventually Collector.stopped will be set
        self.collector.stopped.wait()
        # Stop our own processing
        self.queue.put(None)
        # Wait at most 10 secs for the above `None` in the queue to be processed
        self.stopped.wait(10)
        self.close()
        self.terminate()
